package net.jolsrv2;

import java.util.*;
import java.net.*;
import java.util.concurrent.*;
import java.util.logging.Logger;

import net.jolsrv2.comm.*;
import net.jolsrv2.constants.*;
import net.jolsrv2.data.*;
import net.jolsrv2.repository.neighbourhood.LinkSet;
import net.jolsrv2.repository.neighbourhood.LinkTuple;
import net.jolsrv2.repository.neighbourhood.MprSelectorSet;
import net.jolsrv2.repository.neighbourhood.MprSet;
import net.jolsrv2.repository.neighbourhood.NaaSet;
import net.jolsrv2.repository.neighbourhood.SymmetricNeighbourSet;
import net.jolsrv2.repository.neighbourhood.TwoHopSet;
import net.jolsrv2.repository.processforward.ReceivedSet;
import net.jolsrv2.repository.topology.AdvertisedNeighbourSet;
import net.jolsrv2.repository.topology.AnsnHistorySet;
import net.jolsrv2.repository.topology.AttachedNetworkSet;
import net.jolsrv2.repository.topology.RoutingSet;
import net.jolsrv2.repository.topology.TopologySet;
import net.jolsrv2.utils.*;


/**
 * The class is responsible for dealing with all things related to the interfaces
 * @author Andreas Schjnhaug
 *
 */
public class Interface {
	
	private static Logger logger = Logger.getLogger(Interface.class.toString());
	
	private int packetSequenceNumber = 0;
	
	private DatagramSocket socket;
	
	private CidrAddress localCAddress;
	private InetAddress broadcastAddress;
	
	//Repositories
	private ReceivedSet receivedSet;
	private Sender sender;
	private Receiver receiver;
	private MprSet mprSet;

	private ArrayBlockingQueue<Message> outgoingMessages;
	private ArrayBlockingQueue<ParsedPacket> parsedPackets;



	
	/**
	 * 
	 * @param multiple whether or not multiple interfaces are used
	 * @param localCAddress the local CIDR address the local CIDR address
	 * @param broadcastAddress the broadcast Address the broadcast address
	 * @param port the port to be used for broadcasts
	 */
	public Interface(boolean multiple, CidrAddress localCAddress, InetAddress broadcastAddress, int port){
		
		this.localCAddress = localCAddress;
		this.broadcastAddress = broadcastAddress;
		
		
		
		/**
		 * The following is a work around for a known bug (Bug ID: 4191980)
		 * in the JVM on Linux systems.
		 * 
		 */		
		try{
			if (multiple)
				socket = new DatagramSocket(port,localCAddress.getAddress());
			else
				socket = new DatagramSocket(port);
			socket.setBroadcast(true);
			
		}
		catch(SocketException e){logger.warning("ERROR (1) in Interface constructor: " + e);}
		catch(SecurityException e){logger.warning("ERROR (2) in Interface constructor: " + e);}
		
		receivedSet = new ReceivedSet();
		mprSet = new MprSet(localCAddress);
			

		outgoingMessages = new ArrayBlockingQueue<Message>(100);
		parsedPackets = new ArrayBlockingQueue<ParsedPacket>(100);
		
		sender = new Sender(this, new InetSocketAddress(broadcastAddress, port));
		receiver = new Receiver(this);
		
		Executor executor = Executors.newFixedThreadPool(2);

		executor.execute(sender);
		executor.execute(receiver);
		
	}
	/**
	 * Creates a new HELLO message and adds it to the queue of outgoing messages
	 *
	 */
	public void generateHello (){
		
		logger.info("Generating HELLO on interface " + localCAddress);
		Message hello = new Message();

		hello.setType(MessageTypes.HELLOV2);
		hello.setSemantics(0);
		hello.setOriginatorAddress(Node.getNode().getOriginatorAddress());
		hello.setHopLimit(1);
		hello.setHopCount(0);
		hello.setSequenceNumber(Node.getNode().getAMessageSequenceNumber());
		
		
		TlvBlock messageTlvBlock = new TlvBlock();
		
		Tlv validityTime = new Tlv(TlvTypes.VALIDITY_TIME);
		validityTime.setSemantics(4); //noindex set
		validityTime.setValue(Mantissa.encode(Values.L_HOLD_TIME.byteValue()));
		messageTlvBlock.add(validityTime);
		
		Tlv intervalTime = new Tlv(TlvTypes.INTERVAL_TIME);
		intervalTime.setSemantics(4); //noindex set
		intervalTime.setValue(Mantissa.encode(Values.HELLO_INTERVAL.byteValue()));
		messageTlvBlock.add(intervalTime);
		
		
		if(!Node.getNode().getWillingness().equals(Values.WILL_DEFAULT)){
			Tlv willingnessTlv = new Tlv(TlvTypes.WILLINGNESS);
			willingnessTlv.setSemantics(4);//noindex set
			willingnessTlv.setValue(Node.getNode().getWillingness());
			messageTlvBlock.add(willingnessTlv);
		}
		
		hello.setMessageTlvBlock(messageTlvBlock);
		
		
		//Making the Local Interface Block
		ArrayList<Interface> allInterfaces = Node.getNode().getInterfaces();
		ArrayList<CidrAddress> allCidrAddresss = new ArrayList<CidrAddress>();
		for(Interface iface : allInterfaces){
			allCidrAddresss.add(iface.getLocalCAddress());
		}
		
		
		TlvBlock libTlvBlock = new TlvBlock();
		
		if(allCidrAddresss.size() > 1){
			//Making OTHER_IF
			Tlv otherIf = new Tlv(TlvTypes.OTHER_IF);
			otherIf.setSemantics(2); // novalue set
			
			otherIf.setIndexStart(1);
			otherIf.setIndexStop(allCidrAddresss.size()-1);
			
			libTlvBlock.add(otherIf);
			
			//Rearranging the address so that this interface's address is first
			allCidrAddresss.remove(localCAddress);
			allCidrAddresss.add(0, localCAddress);
			
		}
		
		AddressBlock libAddressBlock = new AddressBlock();
		for(CidrAddress CidrAddress : allCidrAddresss)
			libAddressBlock.add(CidrAddress.getAddress());
		
		hello.addAddressBlock(libAddressBlock);
		
		//Making PREFIX tlv
		
		TlvBlock prefixTlvBlock = new TlvBlock();
		for(CidrAddress cidrAddress : allCidrAddresss){
			if(cidrAddress.getAddress().getAddress().length != (cidrAddress.getPrefixLength() / 8)){
				//Tlv prefix
				logger.finest("TODO: make prefix TLV"); //TODO
			}
		}
		hello.addTlvBlock(prefixTlvBlock);
		
		
		
		
		//Gather the links
		
		LinkSet linkSet = LinkSet.getLinkSet();
		
		
		if(linkSet.contains(localCAddress)){
			
			
			ArrayList<CidrAddress> heard = new ArrayList<CidrAddress>();
			ArrayList<CidrAddress> symmetric = new ArrayList<CidrAddress>();
			ArrayList<CidrAddress> lost = new ArrayList<CidrAddress>();
			
				
			for(LinkTuple linkTuple : linkSet.getTuples(localCAddress)){
				
				Values linkStatus = linkTuple.getStatus();
				
				
				if (linkStatus.equals(Values.HEARD)){
					heard.add(linkTuple.getNeighbourCAddress());							
				}
				else if (linkStatus.equals(Values.SYMMETRIC)){
					symmetric.add(linkTuple.getNeighbourCAddress());							
				}
				else if(linkStatus.equals(Values.LOST)){
					lost.add(linkTuple.getNeighbourCAddress());				
				}		
				
			}
	
			//Making Tlvs
			AddressBlock addressBlock = new AddressBlock();
			TlvBlock tlvBlock = new TlvBlock();
			
			
			if(!symmetric.isEmpty()){
				logger.fine("- Making SYMMETRIC tlv concerning " + symmetric.size() + " links");
				
				
				Tlv lStatus = new Tlv(TlvTypes.LINK_STATUS);
				lStatus.setValue(Values.SYMMETRIC);
				//Valid for one address
				if(lost.size() == 1){
					lStatus.setSemantics(8); //singleindex set
					lStatus.setIndexStart(0);
				}
				//Valid for all addresses
				else if(symmetric.isEmpty() && heard.isEmpty()){
					lStatus.setSemantics(4); //noindex set
				}
				//Valid for an interval of addresses
				else{
					lStatus.setIndexStart(0);
					lStatus.setIndexStop(symmetric.size() - 1);
				}
				
				for(CidrAddress symmetricAddress : symmetric)
					addressBlock.add(symmetricAddress.getAddress());
				tlvBlock.add(lStatus);

				//See if some of the neighbours has been selected as MPR
				ArrayList<InetAddress> mprs = mprSet.getAddresses();
				
				if(mprs.size() != 0){
					int start = -1;
					int counter = 0;
					for(InetAddress mprAddress : mprs){
						for(int i = 0; i < symmetric.size(); i++){
							if(mprAddress.equals(symmetric.get(i).getAddress())){
								if(start == -1)
									start = i;
								counter++;
							}
							else if(start != -1){
								Tlv mprTlv = new Tlv(TlvTypes.MPR);
								mprTlv.setIndexStart(start);
								if(counter == 0){//single MPR
									mprTlv.setSemantics(9); //singleindex & novalue
								}
								else{//Interval of MPRs
									mprTlv.setSemantics(1); //novalue
									mprTlv.setIndexStop(start + counter);
									
								}
								tlvBlock.add(mprTlv);
								start = -1;
								counter = 0;
							}
						}
					}
				}
			}
			
			
			if(!heard.isEmpty()){
				logger.fine("- Making HEARD tlv concerning " + heard.size() + " links");
				
				Tlv lStatus = new Tlv(TlvTypes.LINK_STATUS);
				lStatus.setValue(Values.HEARD);
				//Valid for one address
				if(lost.size() == 1){
					lStatus.setSemantics(8); //singleindex set
					lStatus.setIndexStart(symmetric.size());
				}
				//Valid for all addresses
				else if(symmetric.isEmpty() && heard.isEmpty()){
					lStatus.setSemantics(4); //noindex set
				}
				//Valid for an interval of addresses
				else{
					lStatus.setIndexStart(symmetric.size());
					lStatus.setIndexStop(symmetric.size() + heard.size() - 1);
				}
				
				for(CidrAddress heardAddress : heard)
					addressBlock.add(heardAddress.getAddress());
				tlvBlock.add(lStatus);
			}
			
			
			if(!lost.isEmpty()){
				logger.fine("- Making LOST tlv concerning " + lost.size() + " links");
				Tlv lStatus = new Tlv(TlvTypes.LINK_STATUS);
				lStatus.setValue(Values.LOST);
				//Valid for one address
				if(lost.size() == 1){
					lStatus.setSemantics(8); //singleindex
					lStatus.setIndexStart(symmetric.size() + heard.size());
				}
				//Valid for all addresses
				else if(symmetric.isEmpty() && heard.isEmpty()){
					lStatus.setSemantics(4); //noindex set
				}
				//Valid for an interval of addresses
				else{
					lStatus.setSemantics(0);
					lStatus.setIndexStart(symmetric.size() + heard.size());
					lStatus.setIndexStop(symmetric.size() + heard.size() + lost.size() - 1);
				}
				
				for(CidrAddress inetAddress : lost)
					addressBlock.add(inetAddress.getAddress());
				
				tlvBlock.add(lStatus);
			}
			
			
			
			
			
			
			
			
			hello.addAddressBlock(addressBlock);
			hello.addTlvBlock(tlvBlock);
		}
		else
			logger.fine("No links in LinkSet");

		
	
		try{
			outgoingMessages.offer(hello);
		}
		catch(Exception e){logger.warning("ERROR in generateHello: " + e);}
	}
	/**
	 * Creates a new TC message and adds it to the queue of outgoing messages
	 *
	 */
	public void generateTc(){
		
		logger.info("Generating TC on interface " + localCAddress);
		

		Message tc = new Message();
		tc.setType(MessageTypes.TCV2);
		tc.setSemantics(0);
		tc.setOriginatorAddress(Node.getNode().getOriginatorAddress());
		tc.setHopLimit(255);
		tc.setHopCount(0);
		tc.setSequenceNumber(Node.getNode().getAMessageSequenceNumber());
		
		TlvBlock messageTlvBlock = new TlvBlock();
		
		Tlv vTime = new Tlv(TlvTypes.VALIDITY_TIME);
		vTime.setSemantics(4);
		vTime.setValue(Mantissa.encode(Values.TC_INTERVAL.byteValue()*3)); //TODO: Value not specified at the time of implementation
		messageTlvBlock.add(vTime);
		
		Tlv intervalTime = new Tlv(TlvTypes.INTERVAL_TIME);
		intervalTime.setSemantics(4);
		intervalTime.setValue(Mantissa.encode(Values.TC_INTERVAL.byteValue()));
		messageTlvBlock.add(intervalTime);
		
		
		Tlv contentSeqTlv = new Tlv(TlvTypes.CONTENT_SEQ_NO);
		contentSeqTlv.setSemantics(4);
		contentSeqTlv.setValue((byte)AdvertisedNeighbourSet.getAdvertisedNeighbourSet().getAnsn());
		messageTlvBlock.add(contentSeqTlv);

		tc.setMessageTlvBlock(messageTlvBlock);
		
		//Creating Local Information Block
		//TODO: also exists in generateHello() & should rearrange??
		ArrayList<Interface> allInterfaces = Node.getNode().getInterfaces();
	
		AddressBlock libAddressBlock = new AddressBlock();
		for(Interface iface : allInterfaces){
			libAddressBlock.add(iface.getLocalCAddress().getAddress());
		}
		tc.addAddressBlock(libAddressBlock);
		tc.addTlvBlock(new TlvBlock()); //TODO PREFIX
		
		
		ArrayList<InetAddress> advertisedNeighbours = AdvertisedNeighbourSet.getAdvertisedNeighbourSet().getAddresses();
		if(!advertisedNeighbours.isEmpty()){
			
			logger.fine("The following neighbours are advertised:");
			
			AddressBlock advertisedNeighboursBlock = new AddressBlock();
			for(InetAddress address : advertisedNeighbours){
				logger.fine(address.toString());
				advertisedNeighboursBlock.add(address);
			}
			
			
			tc.addAddressBlock(advertisedNeighboursBlock);
			tc.addTlvBlock(new TlvBlock());//TODO fix this if multiple ifaces
		}
		
		
		
		ArrayList<InetAddress> gateways = Node.getNode().getGateways();
		if(!gateways.isEmpty()){
			
			logger.fine("The following gateways are announced:");
			
			
			AddressBlock gatewayAddressBlock = new AddressBlock();
			
			for(InetAddress gateway : gateways){
				logger.fine(gateway.toString());
				gatewayAddressBlock.add(gateway);
			}
			
			
			Tlv gatewayTlv = new Tlv(TlvTypes.GATEWAY);
			gatewayTlv.setSemantics(5);
			
			
			TlvBlock gatewayTlvBlock = new TlvBlock();
			gatewayTlvBlock.add(gatewayTlv);
			
			tc.addAddressBlock(gatewayAddressBlock);
			tc.addTlvBlock(gatewayTlvBlock);
			
		}
		
		try{
			outgoingMessages.offer(tc);
		}
		catch(Exception e){logger.warning("ERROR in Interface.generateTc(): " + e);}
		
		
		
		
	}
	

	/**
	 * Returns the local CIDR address
	 * @return the CIDR block attached to the interface
	 */
	public CidrAddress getLocalCAddress(){
		return localCAddress;
	}
	
	/**
	 * Takes one or more messages from the outgoing queue in a FIFO manner
	 * @return the messages
	 */
	public ArrayList<Message> takeFromOutgoingMessages(){
		ArrayList<Message> messages = new ArrayList<Message>();
		
		//Tries to return as many as available
		try{
			messages.add(outgoingMessages.take());
			
			for(int i = 0 ; i < outgoingMessages.size() ; i++)
				messages.add(outgoingMessages.take());
		}
		catch(InterruptedException e){logger.warning("ERROR in takeFromOutgoingMessages: " + e);}

		return messages;
	}
	/**
	 * Populates the MPR set on the Interface
	 *
	 */
	public void populateMprSet(){
		mprSet.populate();
	}
	/**
	 * Return the MPRs on the Interface
	 * @return the IP addresses of the MPRs
	 */
	public ArrayList<InetAddress> getMprs(){
		return mprSet.getAddresses();
	}
	
	/**
	 * Puts a new message in the outgoing messages queue.
	 * @param message to be put in the queue
	 */
	public void offerToOutgoingMessages(Message message){
		
		outgoingMessages.offer(message);
	}
	
	/**
	 * Puts a new parsed packet in the parsed packets queue
	 * @param parsedPacket to be put in the queue
	 */
	public void offerToParsedPackets(ParsedPacket parsedPacket){
		parsedPackets.offer(parsedPacket);
	}
	
	
	
	/**
	 * Returns the next packet sequence number. Wrap around is handled
	 * @return the next packet sequence number
	 */
	public int getAPacketSequenceNumber(){
		if(packetSequenceNumber < 65535)
			return packetSequenceNumber++;
		else{
			packetSequenceNumber = 0;
			return packetSequenceNumber;
		}
	}
	
	/**
	 * Purges the received set
	 *
	 */
	public void purgeReceivedSet(){
		receivedSet.purge();
	}
	
	/**
	 * Returns the datagram socket
	 * @return the datagram socket
	 */
	public DatagramSocket getSocket(){
		return socket;
	}
	
	/*
	 * Returns the broadcastt address
	 * 
	 */
	public InetAddress getBroadcastAddress(){
		return broadcastAddress;
	}
	
	
	
	/**
	 * The methods takes an incoming message, populates the repositories with it
	 * and checks if it is to be forwarded
	 */
	public void processParsedPackets(){
		
		if(!parsedPackets.isEmpty()){
		
			ParsedPacket parsedPacket = null;
			try{
				parsedPacket = parsedPackets.take();	
			}
			
			
			catch(Exception e){logger.warning(e.toString());}
			

			logger.info("Populating repositories");
			
			
			ArrayList<Message> messages = parsedPacket.getMessages();
			CidrAddress sendingCAddress = new CidrAddress(parsedPacket.getSendingAddress());
			
			for(Message message : messages){
			
				long validityTime = -1;
				long intervalTime = -1;
				int ansn = -1;
				Values willingness = Values.WILL_DEFAULT;
				
				for(Tlv tlv : message.getMessageTlvs()){
					if(tlv.getType().equals(TlvTypes.VALIDITY_TIME) && tlv.getValue().length == 1){
						validityTime = Mantissa.decode(tlv.getValue()[0]);
					}
					else if(tlv.getType().equals(TlvTypes.INTERVAL_TIME) && tlv.getValue().length == 1)
						intervalTime = Mantissa.decode(tlv.getValue()[0]);
					else if(tlv.getType().equals(TlvTypes.CONTENT_SEQ_NO) && tlv.getValue().length == 1)
						ansn = tlv.getValue()[0];
					else if(tlv.getType().equals(TlvTypes.WILLINGNESS) && tlv.getValue().length == 1){
						for(Values value : Values.values())
							if(value.byteValue() == tlv.getValue()[0]){
								willingness = value;
								break;
							}
					}
					
				}
				
				
			
				boolean changeInSets = false;
				
				
				//Transforming all addresses to CIDR blocks
				ArrayList<AddressBlock> addressBlocks = message.getAddressBlocks();
				ArrayList<TlvBlock> tlvBlocks = message.getTlvBlocks();
				
				//Local information block
				ArrayList<CidrAddress> libCAddresses = null;
				
				
				//Looping through all the address blocks and populating the sets accordingly
				for(int i = 0; i < addressBlocks.size() ; i++){
					
					//Transforms the addresses in a block to CIDR addresses (if PREFIX TLVs are found)
					ArrayList<CidrAddress> cidrAddresses = new ArrayList<CidrAddress>();
					
					
					//Creating new CIDR addresses from InetAddresses
					for(InetAddress inetAddress : addressBlocks.get(i).getAddresses()){
						cidrAddresses.add(new CidrAddress(inetAddress));
					}
					
					//Checking to see if there are PREFIX TLVs associated with this address block
					for(Tlv tlv : tlvBlocks.get(i).getTlvs()){
							if(tlv.getType().equals(TlvTypes.PREFIX_LENGTH) && i >= tlv.getIndexStart() && i < tlv.getIndexStop()) //TODO may not cover all cases
								logger.warning("TODO: Interface.processParsedPackets: prefix found, must be saved");
								
						
					}

					if(i==0){
						//Saving the LIB CIDR addresses for later use
						libCAddresses = cidrAddresses;
						//Populating neighbourhood address association set
						
						
						
						//Populating link set
						if(message.getType().equals(MessageTypes.HELLOV2) && LinkSet.getLinkSet().populate1(localCAddress, sendingCAddress, validityTime)); //Returns a detection of change in the linkSet
							changeInSets = true;
						
						//TODO Must return changeinsets for routing table
						NaaSet.getNaaSet().populate(intervalTime, cidrAddresses);
						
					}
					
					//------------------------
					//HELLO message processing
					//------------------------
					else if(message.getType().equals(MessageTypes.HELLOV2)){
						
						
						
						//Populating link set
						LinkSet.getLinkSet().populate2(localCAddress, sendingCAddress, validityTime, intervalTime, willingness, cidrAddresses, tlvBlocks.get(i).getTlvs()); //Returns a detection of change in the linkSet
						//	changeInSets = true;
						
						//Populating the symmetric neighbour set
						SymmetricNeighbourSet.getSymmetricNeighbourSet().populate(localCAddress, libCAddresses, cidrAddresses, tlvBlocks.get(i).getTlvs(), validityTime);
						
						
						//Populating two hop set
						if(TwoHopSet.getTwoHopSet().populate(localCAddress, sendingCAddress, validityTime, cidrAddresses, tlvBlocks.get(i).getTlvs()));
							changeInSets = true;
								
							
						//Populating the MPR selector set
						if(MprSelectorSet.getMprSelectorSet().populate(localCAddress, libCAddresses, cidrAddresses, tlvBlocks.get(i).getTlvs(), validityTime)){
							//Populating derived sets
							Node.getNode().populateDerivedSets();
						}
					}
					
					
					
					
					//---------------------
					//TC message processing
					//---------------------
					else if(message.getType().equals(MessageTypes.TCV2)){
						
						if(AnsnHistorySet.getAnsnHistorySet().populate(message.getOriginatorAddress(),ansn, validityTime)){
							if(TopologySet.getTopologySet().update(libCAddresses, addressBlocks.get(i).getAddresses(), tlvBlocks.get(i).getTlvs(), validityTime, ansn))
								changeInSets = true;
							
							AttachedNetworkSet.getAttachedNetworkSet().populate(libCAddresses,cidrAddresses,tlvBlocks.get(i).getTlvs(),ansn,validityTime);
							
						}
						
					}
						
					//--------------
					//  Message considered for forwarding
					//--------------
					if(message.getHopLimit() > 0 && !message.getType().equals(MessageTypes.HELLOV2)){ 
						//Message is considered for forwarding
						
						
						//TODO Check to see if there is prefix information regarding the sending address
						
						
						//Check to see if there is a symmetric link
						if(LinkSet.getLinkSet().contains(localCAddress, sendingCAddress, Values.SYMMETRIC)){
							
							//Check to see if message has already been received
							if (receivedSet.contains(message.getType(), message.getOriginatorAddress(), message.getSequenceNumber())){
								//Skip the message
								logger.fine("    - Message already received, will not forward..");
								//buffer.position(buffer.position() + msgSize - ipLength - 8);
							}
							else{
								receivedSet.addTuple(message.getType(), message.getOriginatorAddress(), message.getSequenceNumber(), System.nanoTime() + 10*(long)1e9); //TODO: R_HOLD_TIME value not specified
							
							}
							
							
							//Forward this message
							
							Node.getNode().forwardMessage(sendingCAddress.getAddress(), message);
							
							
						}					
					}				
				}
				
				if(changeInSets){
					
					for(Interface i : Node.getNode().getInterfaces())
						i.populateMprSet();
					
					
					RoutingSet.getRoutingSet().calculate();
				}
				
				
				logger.info("---------------------------");
			}
		}
	}
}

