package net.jolsrv2;
import java.util.*;
import java.util.concurrent.Executors;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;
import java.io.File;
import java.io.FileNotFoundException;
import java.net.*;


import net.jolsrv2.constants.*;
import net.jolsrv2.data.*;
import net.jolsrv2.repository.neighbourhood.LinkSet;
import net.jolsrv2.repository.neighbourhood.MprSelectorSet;
import net.jolsrv2.repository.neighbourhood.NaaSet;
import net.jolsrv2.repository.neighbourhood.SymmetricNeighbourSet;
import net.jolsrv2.repository.neighbourhood.TwoHopSet;
import net.jolsrv2.repository.processforward.ForwardedSet;
import net.jolsrv2.repository.processforward.ProcessedSet;
import net.jolsrv2.repository.processforward.RelaySet;
import net.jolsrv2.repository.topology.AdvertisedNeighbourSet;

/**
 * The main class
 * @author Andreas Schjnhaug
 *
 */
public final class Node {
	
	
	private static Logger logger = Logger.getLogger(Node.class.toString());
	
	
	private static Node ref;

	private int port;
	private boolean flagHello;
	private boolean flagTc;
	
	private ArrayList<InetAddress> mprUnionSet;
	
	private int ipVersion;
	 
	private Values willingness;
	private InetAddress originatorAddress;
	
	private ArrayList<Interface> interfaces;
	private ArrayList<InetAddress> gateways;
	

	
	private int messageSequenceNumber = 0;
	
	/**
	 * Implementation of the singleton design pattern
	 * @return the node
	 */
	public static synchronized Node getNode(){
		if(ref == null){
			try{
				ref = new Node();
			}
			catch(Exception e){logger.severe(e.toString());}
		}
		return ref;
	}
	
	/**
	 * Implementation of the singleton design pattern. This method is used to
	 * ensure that the class can not be cloned.
	 */
	public Object clone() throws CloneNotSupportedException{
		throw new CloneNotSupportedException();
	}
	
	
	/**
	 Constructor

	*/
	private Node() {

		logger.info("jOLSRv2");
	
		port = 50000;
		
		mprUnionSet = new ArrayList<InetAddress>();

		
		interfaces = new ArrayList<Interface>();
		gateways = new ArrayList<InetAddress>();

		
		//Check if the Operating system is supported

		if(!System.getProperty("os.name").equals("Windows XP"))
			logger.warning("Please note that routing table management is currently not supported for " + System.getProperty("os.name"));
		
				
		ArrayList<String> names = new ArrayList<String>();
		ArrayList<Integer> prefixes = new ArrayList<Integer>();;
		ArrayList<InetAddress> broadcastAddresses = new ArrayList<InetAddress>();
		ipVersion = 4;
		try{
			Scanner scanner = new Scanner(new File("config.txt"));
		
			//Node settings
			if(scanner.hasNext("IPVERSION")){
				scanner.findInLine("IPVERSION");
				ipVersion = scanner.nextInt();
				if(ipVersion != 4 && ipVersion != 6){
					logger.severe("IPVERSION must be 4 or 6, please check config.txt");
					System.exit(1);
				}
			}
			else{
				logger.severe("IPVERSION missing in config.txt");
				System.exit(1);
			}
			
			
			
			if(scanner.hasNext("WILLINGNESS")){
				scanner.findInLine("WILLINGNESS");
				String will = scanner.next();
				if(will.equals("NEVER"))
					willingness = Values.WILL_NEVER;
				else if(will.equals("LOW"))
					willingness = Values.WILL_LOW;
				else if(will.equals("DEFAULT"))
					willingness = Values.WILL_DEFAULT;
				else if(will.equals("HIGH"))
					willingness = Values.WILL_HIGH;
				else if(will.equals("ALWAYS"))
					willingness = Values.WILL_ALWAYS;
				else{
					willingness = Values.WILL_DEFAULT;
					logger.warning("Willingess not recognised, using WILL_DEFAULT");
				}

			}
			else{
				logger.severe("WILLINGNESS missing in config.txt");
				System.exit(1);
			}
			
			
			
			while(scanner.hasNext("GATEWAY")){
				scanner.findInLine("GATEWAY");
				try{
					gateways.add(InetAddress.getByName(scanner.next()));
				}
				catch(UnknownHostException e){logger.warning("Error in gateway address. The gateway will not be announced.");}
			}
			
			
			
			//Interface settings
			
			while(scanner.hasNext("INTERFACE")){
				scanner.findInLine("INTERFACE");
				names.add(scanner.next());
				
			}
			
			
			while(scanner.hasNext("PREFIX")){
				scanner.findInLine("PREFIX");
				prefixes.add(scanner.nextInt());
				
			}
			
			while(scanner.hasNext("BROADCAST")){
				scanner.findInLine("BROADCAST");
				try{
					broadcastAddresses.add(InetAddress.getByName(scanner.next()));
				}
				catch(UnknownHostException e){
					logger.severe("Error in broadcast address, please check your config file.");
					System.exit(1);
					}
			}
			
			
		}
		catch(FileNotFoundException e){
			logger.severe("config.txt file not found.");
			System.exit(1);
		}

		try{
			for(int i = 0 ; i < names.size(); i++){

    			NetworkInterface ni = NetworkInterface.getByName(names.get(i));
    			if(ni == null){
    				logger.severe("Cannot find the interface names " + names.get(i) + ", please check your config file");
    				System.exit(1);
    			}
    			
    			Enumeration ipAddresses = ni.getInetAddresses();
    			
				while(ipAddresses.hasMoreElements()){
					InetAddress localAddress = (InetAddress)ipAddresses.nextElement();
					
					if(originatorAddress == null){
						if (localAddress.getClass() == Inet4Address.class && ipVersion == 4){
							originatorAddress = localAddress;
							
						}
					}
					
					if(localAddress.getClass() == Inet4Address.class && ipVersion == 4){
						logger.fine("- Using IPv4 address " + localAddress + " on " + ni.getDisplayName());
						interfaces.add(new Interface(names.size() != 1 ? true : false,new CidrAddress(localAddress,prefixes.get(i)),broadcastAddresses.get(i),port));
				
					  
					}
					else if(localAddress.getClass() == Inet6Address.class && ipVersion == 6){
						logger.finest("    TODOFound IPv6 address: " + localAddress);
					}
					
				}
    		}
		}
		catch (SocketException e){
			logger.severe("SocketException in Node: " + e);
			System.exit(1);
			}

		if(interfaces.isEmpty()){
			logger.severe("Could not find a suitable interface, please check your config file.");
			System.exit(1);
		}
		
		logger.config("* Operating system: " + System.getProperty("os.name"));
		logger.config("* Port:        " + port);
		for(int i = 0; i < interfaces.size(); i++){
			logger.config("* Interface:   " + interfaces.get(i).getLocalCAddress() + "     Broadcast: " + broadcastAddresses.get(i));
		}
		logger.config("* Originator address: " + originatorAddress);
		logger.config("* Willingness: " + willingness);
		if(!gateways.isEmpty())
			logger.config("* Gateways: " + gateways);
		
		
		ScheduledExecutorService service = Executors.newScheduledThreadPool(2);
		
		Random random = new Random();
		
		final Runnable hello = new Runnable(){
			public void run(){
				flagHello = true;//generate Hello 
			}
		};
		try{
			service.scheduleAtFixedRate(hello,(long)0,(long)2000 - random.nextInt(500),TimeUnit.MILLISECONDS);
		}
		catch(RejectedExecutionException e){logger.warning(e.toString());}
		catch(NullPointerException e){logger.warning(e.toString());}
		catch(IllegalArgumentException e){logger.warning(e.toString());}
		
		
		final Runnable tc = new Runnable(){
			public void run(){
				if(!(AdvertisedNeighbourSet.getAdvertisedNeighbourSet().getAddresses().isEmpty() && gateways.isEmpty()))
					flagTc = true;//generate TC
			}
		};
		try{
			service.scheduleAtFixedRate(tc,(long)0,(long)5000 - random.nextInt(500),TimeUnit.MILLISECONDS);	
		}
		catch(RejectedExecutionException e){logger.warning(e.toString());}
		catch(NullPointerException e){logger.warning(e.toString());}
		catch(IllegalArgumentException e){logger.warning(e.toString());}
		
		
		
		
		ScheduledExecutorService executor2 = Executors.newSingleThreadScheduledExecutor();
		final Runnable purger = new Runnable(){
			public void run(){
				purgeRepositories();

				if(flagHello){
					for (Interface i : interfaces)
						i.generateHello();
					flagHello = false;
				}
				
				if(flagTc){
					for (Interface i : interfaces)
						i.generateTc();
					flagTc = false;
				}
				
				for (Interface i : interfaces)
					i.processParsedPackets();	
			}
		};
		try{
			executor2.scheduleAtFixedRate(purger,(long)0,(long)500,TimeUnit.MILLISECONDS);
		}
		catch(RejectedExecutionException e){logger.warning(e.toString());}
		catch(NullPointerException e){logger.warning(e.toString());}
		catch(IllegalArgumentException e){logger.warning(e.toString());}
			
	}
	
	/**
	 * The method return the MPRs 
	 *
	 * @return ArrayList of InetAddresses
	 */
	public ArrayList<InetAddress> getMprs(){
		
		mprUnionSet.clear();
		for(Interface i : interfaces){
			ArrayList<InetAddress> mprs = i.getMprs();
			for(InetAddress ia : mprs){
				if (!mprUnionSet.contains(ia))
					mprUnionSet.add(ia);
			}
			
		}
		
		return mprUnionSet;
	}
	
	/**
	 * Returns the Interfaces
	 * @return an ArrayList of Interfaces
	 */
	public ArrayList<Interface> getInterfaces(){
		return interfaces;
	}
	
	/**
	 * The method purges the following repositories:
	 * <ul>
	 * <li>LinkSet
	 * <li>TwoHopSet
	 * <li>MprSelectorSet
	 * <li>NaaSet
	 * <li>ProcessedSet
	 * <li>ForwardedSet
	 * <li>ReceivedSet
	 * </ul>
	 * Notes:
	 * Changes in the linkSet or TwoHop will cause a repopulation of the MprSet
	 * Changes in the MprSelectorSet will cause a repopulation of the Relay Set and AdvertisedNeighbourSet
	 *
	 */	
	private void purgeRepositories(){
		
		boolean changeInSets = false;
		if(LinkSet.getLinkSet().purge())
			changeInSets = true;
		
		if(SymmetricNeighbourSet.getSymmetricNeighbourSet().purge())
			changeInSets = true;
		
		if(TwoHopSet.getTwoHopSet().purge())
			changeInSets = true;
				
		if (changeInSets)
			for(Interface i : interfaces)
				i.populateMprSet();
		
		
		if(MprSelectorSet.getMprSelectorSet().purge()){
			populateDerivedSets();
		}
		
		
		NaaSet.getNaaSet().purge();
		
		//TODO: purge Topolgy or purge when creating routing table?

		
		ProcessedSet.getProcessedSet().purge();
		ForwardedSet.getForwardedSet().purge();
		
		for(Interface i: interfaces){
			i.purgeReceivedSet();
		}
		
	}
	
	/**
	 * The method populates the derived sets, meaning:
	 * <ul>
	 * <li>RelaySet
	 * <li>AdvertisedNeighbourSet
	 * </ul>
	 *
	 */
	public void populateDerivedSets(){
		RelaySet.getRelaySet().clear();
		ArrayList<CidrAddress> mprSelectorCAddresses = MprSelectorSet.getMprSelectorSet().getCAddresses();
		for(CidrAddress cAddress : mprSelectorCAddresses){
			RelaySet.getRelaySet().add(cAddress.getAddress());
		}
		
		AdvertisedNeighbourSet.getAdvertisedNeighbourSet().populate();
	
	}
	

	/**
	 * The method returns the next message sequence number to be used. It handles wrap around
	 * @return next message sequence number to be used
	 */
	public int getAMessageSequenceNumber(){
		if(messageSequenceNumber < 65535)
			return messageSequenceNumber++;
		else{
			messageSequenceNumber = 0;
			return messageSequenceNumber;
		}
	}
	
	
	
	/**
	 * The method return a node's willingness
	 * @return the node's willingness
	 */
	public Values getWillingness(){
		return willingness;
	}
	
	
	/**
	 * The method evaluates if the incoming message should be forwarded,
	 * and acts accordingly. If the message is accepted for forwarding,
	 * it is offered to the outgoing message queue on all interfaces.
	 * @param 	sourceAddress 	the address of the interface receiving the message
	 * @param 	message 		the message
	 */
	public void forwardMessage(InetAddress sourceAddress, Message message){
		
		if (!ForwardedSet.getForwardedSet().contains(message.getType(), message.getOriginatorAddress(), message.getSequenceNumber())){
			
			if(RelaySet.getRelaySet().contains(sourceAddress)){
				ForwardedSet.getForwardedSet().addTuple(message.getType(), message.getOriginatorAddress(), message.getSequenceNumber(), System.nanoTime() + 10); //TODO: 10 should be F_HOLD_TIME
				message.setHopLimit(message.getHopLimit()-1);
				message.setHopCount(message.getHopCount()+1);
				logger.info("Forwarding message on all interfaces");
				for(Interface i : interfaces)
					i.offerToOutgoingMessages(message);
				
			}
			
		}
	}
	
	/**
	 * The moethod gets the address used as originator address on the node
	 * @return InetAddress originatorAddress
	 */
	public InetAddress getOriginatorAddress(){
		return originatorAddress;
	}
	
	/**
	 * The main method.
	 * @param args
	 */
	public static void main(String []args)
	{
		Node.getNode();
	}
	
	/**
	 * The method gets the IP version used (either 4 or 6).
	 * @return int ipVersion
	 */
	public int getIpVersion(){
		return ipVersion;
	}

	/**
	 * The method gets the gateways attached to the node.
	 * 
	 * @return ArrayList<InetAddress> gateways
	 */
	public ArrayList<InetAddress> getGateways(){
		return gateways;
	}
}