package net.jolsrv2.data;

import java.net.*;
import java.util.*;
import java.util.logging.Logger;
import java.nio.*;

import net.jolsrv2.Node;
/**
 * An address is specified as a sequence of octets of the form head:mid:
 * tail. An address block is an ordered set of addresses sharing the
 * same head and tail, and having individual mids.
 * 
 * [address block] is defined by:
 *     [address-block] = [num-addr]
 *     [head-octet]
 *     [head]?
 *     [tail-octet]?
 *     [tail]?
 *     [mid]*
 * 
 * The class implements the address block.
 * 
 * @author Andreas Schjnhaug
 *
 */
public class AddressBlock{
	
	private static Logger logger = Logger.getLogger(AddressBlock.class.toString());
	
	private int headOctet; //byte
	private byte[] head;
	private int tailOctet; //byte
	private byte[] tail; //byte
	private int midLength;
	private byte[] mid;
  
	private ArrayList<InetAddress> addresses;
  
	/**
	 * Constructor
	 *
	 */
	public AddressBlock(){
		addresses = new ArrayList<InetAddress>();
	}
  
	
	/**
	 * The method adds an IP address to the address block
	 * @param address
	 */
	public void add(InetAddress address){
		addresses.add(address);
	
	}
  
	/**
	 * Calculates the head and tails of multiple addresses
     *
     */
	public void calculate(){
		
		int ipVersion = Node.getNode().getIpVersion();
		
		//finding head and tail split
		InetAddress first = addresses.get(0);
		int headSplit = 0;
		int tailSplit = first.getAddress().length - 1;
		byte bHead = first.getAddress()[headSplit];
		byte bTail = first.getAddress()[tailSplit];
		boolean foundHead = false;
		boolean foundTail = false;
		  
		if(addresses.size() != 1)
			while(!foundTail || !foundHead){
		  
				for(int i = 1; i < addresses.size(); i++){
					if(!foundHead && bHead != addresses.get(i).getAddress()[headSplit]){
				  
						foundHead = true;
					    if(foundTail)
					    	break;
					}
			  
					if(!foundTail && bTail != addresses.get(i).getAddress()[tailSplit]){
						  
						foundTail = true;
						if(foundHead)
							break;
					}
				}
		  
		  
				if(!foundHead)
					bHead = first.getAddress()[++headSplit];
				if(!foundTail)
					bTail = first.getAddress()[--tailSplit];
			}		  
		
		//Four cases
		// 1) mid
		// 2) head:mid
		// 3) mid:tail
		// 4) head:mid:tail
		 
		//Case 1 - mid
		if(headSplit == 0 && tailSplit == ipVersion - 1){
			logger.finest("mid");
			  
			//Puts all address consecutevily in the head
			headOctet = 1;
			midLength = ipVersion;
			mid = getMid(ipVersion * addresses.size(), 0, ipVersion - 1);
 
		}
		//Case 2 - head:mid
		else if(headSplit > 0 && tailSplit == ipVersion - 1){
			logger.finest("head:mid");
			  
			headOctet = headSplit << 1;
			headOctet += 1; //notail flag
			head = new byte[headSplit];
			for(int i = 0; i < headSplit ; i++){
				head[i] = first.getAddress()[i];
			}
			  
			midLength = ipVersion - head.length;
			mid = getMid(midLength * addresses.size(), headSplit, ipVersion - 1);  
		}
		  
		//Case 3 - mid:tail
		else if(headSplit == 0 && tailSplit < ipVersion - 1){
			logger.finest("mid:tail");
			  
			headOctet = 0;
			  
			tailOctet = (ipVersion - tailSplit - 1) << 1;
			tail = new byte[ipVersion - tailSplit - 1];
			int x = 0;
			for(int i = tailSplit + 1; i < ipVersion; i++){
				tail[x++] = first.getAddress()[i];
			}
			  
			midLength = ipVersion - tail.length;
			mid = getMid(midLength * addresses.size(), 0, tailSplit);
			  
			  
		  
		}
		//Case 4 - head:mid:tail
		else{
			logger.finest("head:mid:tail");
			
			//head
			headOctet = headSplit << 1;
			head = new byte[headSplit];
			for(int i = 0; i < headSplit ; i++){
				head[i] = first.getAddress()[i];
			}
			
			//tail
			tailOctet = (ipVersion - tailSplit - 1) << 1;
			tail = new byte[ipVersion - tailSplit - 1];
			int x = 0;
			for(int i = tailSplit + 1; i < ipVersion; i++){
				tail[x++] = first.getAddress()[i];
			}
			
			//mid
			midLength = ipVersion - head.length - tail.length; 
			mid = getMid(midLength * addresses.size(), headSplit, tailSplit);
		}
			  
	}

	/**
     * Returns the IP addresses
     * @return the addresses contained in the address block
     */
    public ArrayList<InetAddress> getAddresses(){
	    return addresses;
    }
    /**
     * Returns the bytes of the address block
     * @return the bytes in the address block
     * @see Message
     */
    public byte[] getBytes(){
		  
		calculate();
			
		int size = 2 + (headOctet >> 1);
		
		if((headOctet & 0x1) == 0 ){
			size += 1 + (tailOctet >> 1);
			
		}
		if(midLength != 0){
			size += mid.length;
		}
		  
		byte[] bytes = new byte[size];
		  
		ByteBuffer byteBuffer = ByteBuffer.wrap(bytes);

		byteBuffer.put((byte)addresses.size());
		byteBuffer.put((byte)headOctet);
		if((headOctet >> 1) != 0 ){
			byteBuffer.put(head);
		}
		  
		if((headOctet & 0x1) == 0){
			byteBuffer.put((byte)tailOctet);
			if((tailOctet & 0x1) == 0){
				byteBuffer.put(tail);
			}
		}
		
		if(midLength != 0)
			byteBuffer.put(mid);
		  
			  
		if(byteBuffer.capacity() != byteBuffer.position())
			logger.severe("CAPACITY != POSITION");
			 
		return bytes;
	  }

    /**
     * The method return the mid part of the addresses
     * 
     * @param length length of the mid section desired
     * @param start where to start the reading of bytes
     * @param stop where to stop the reading of bytes
     * @return bytes containing the mid section
     */
    private byte[] getMid(int length, int start, int stop){

		byte[] bytes = new byte[length];
		
		int pointer = 0;
		for(InetAddress address : addresses){
			byte[] addressBytes = address.getAddress();
			for(int i = start; i <= stop; i++){
				bytes[pointer++] = addressBytes[i];
			}
		}
		return bytes;
	}
}

