package net.jolsrv2.data;

import java.util.ArrayList;
import java.util.logging.Logger;
import java.net.InetAddress;
import java.nio.*;

import net.jolsrv2.constants.*;
/**
 * The class implements the messages, as specified in OLSRv2.
 * @author Andreas Schjnhaug
 *
 */
public class Message {
	
	private static Logger logger = Logger.getLogger(Message.class.toString());
	
	private MessageTypes type;
	private int semantics;
	private InetAddress originatorAddress;
	private int hopLimit;
	private int hopCount;
	private int sequenceNumber;
	 
	private TlvBlock messageTlvBlock;
	private ArrayList<AddressBlock> addressBlocks;
	private ArrayList<TlvBlock> tlvBlocks;
	 
	private byte[] padding;
	 
	
	 
	  /**
	   * Constructor
	   *
	   */
	 
	public Message(){
		messageTlvBlock = new TlvBlock();
		addressBlocks = new ArrayList<AddressBlock>();
		tlvBlocks = new ArrayList<TlvBlock>();
		 
	}
	 
	/**
	 * Returns the message type
	 * @return the message type
	 */
	public MessageTypes getType(){
		return type;
	}
	
	/**
	 * Sets the message type
	 * @param type the message type
	 */
	public void setType (MessageTypes type){
		this.type = type;
		  
	}
	
	/**
	 * Return the address blocks
	 * @return the address blocks
	 */
	public ArrayList<AddressBlock> getAddressBlocks(){
		return addressBlocks;
	}
	
	/**
	 * Returns the TLV blocks
	 * @return the TLV blocks
	 */
	public ArrayList<TlvBlock> getTlvBlocks(){
		return tlvBlocks;
	}
	
	/**
	 * Return the message TLVs
	 * @return the message TLVs
	 */
	public ArrayList<Tlv> getMessageTlvs(){
		return messageTlvBlock.getTlvs();
	}
	
	/**
	 * Returns the originator IP address
	 * @return the originator IP address
	 */
	public InetAddress getOriginatorAddress(){
		return originatorAddress;
	}

	
	 /**
	  * Sets the originator address
	  * @param inetAddress the originator IP address
	  */
	 public void setOriginatorAddress(InetAddress inetAddress){
		 originatorAddress = inetAddress;
	 }
	
	
	/**
	 * Sets the message TLV block
	 * @param messageTlvBlock the message TLV block
	 */
	public void setMessageTlvBlock (TlvBlock messageTlvBlock){
		this.messageTlvBlock = messageTlvBlock;
	}	 
	 
	/**
	 * Adds an address block to the message
	 * @param addressBlock the address block
	 */
	public void addAddressBlock(AddressBlock addressBlock){
		addressBlocks.add(addressBlock);
	}
	 /**
	  * Adds a TLV block to the message
	  * @param tlvBlock the TLV block
	  */
	 public void addTlvBlock(TlvBlock tlvBlock){
		 tlvBlocks.add(tlvBlock);
	 }
	 
	 /**
	  * Sets the message semantics
	  * @param semantics the message semantics
	  */
	 public void setSemantics(int semantics){
		 this.semantics = semantics;
	 }
	 

	 /**
	  * Returns the hop limit
	  * @return the hop limit
	  */
	 public int getHopLimit(){
		 return hopLimit;
	 }
	 
	 /**
	  * Sets the hop limit
	  * @param hopLimit the hop limit
	  */
	 public void setHopLimit(int hopLimit){
		 this.hopLimit = hopLimit;
	 }
	 
	 /**
	  * Returns the message sequence number
	  * @return the message sequence number
	  */
	 public int getSequenceNumber(){
		 return sequenceNumber;
	 }
	 
	 /*
	  * Sets the message sequence number
	  * @param sequenceNumber the message sequence numberthe message sequence number
	  */
	  
	 public void setSequenceNumber(int sequenceNumber){
		 this.sequenceNumber = sequenceNumber;
	 }
	 
	 /**
	  * Gets the hop count
	  * @return the hop count
	  */
	 public int getHopCount(){
		 return hopCount;
	 }
	 
	 /**
	  * Sets the hop count
	  * @param hopCount the hop count
	  */
	 public void setHopCount(int hopCount){
		 this.hopCount = (byte)hopCount;
	 }
	 
	 /**
	  * Returns the bytes of the message
	  * @return the bytes of the message
	  */
	 public byte[] getBytes (){
		 
		 byte[] messageTlvBytes = messageTlvBlock.getBytes();
		 int size = 8 + originatorAddress.getAddress().length + messageTlvBytes.length;

		 if(addressBlocks.size() != tlvBlocks.size())
			 logger.warning("Not an even number of address and tlv blocks");
		 
		 ArrayList<byte[]> addressBytes = new ArrayList<byte[]>();
		 for(AddressBlock addressBlock : addressBlocks){
			 byte[] addressB = addressBlock.getBytes();
			 addressBytes.add(addressB);
			 size += addressB.length;
		 }
		 
		 ArrayList<byte[]> tlvBytes = new ArrayList<byte[]>();
		 for(TlvBlock tlvBlock : tlvBlocks){
			 byte[] tlvB = tlvBlock.getBytes();
			 tlvBytes.add(tlvB);
			 size += tlvB.length;
			 
		 }
		 
		 //padding		 
		 if (size%4 != 0){
			 padding = new byte[4 - size%4];
			 size += padding.length;
			 for(int i = 0; i < padding.length; i++)
				 padding[i] = 0;
				 
		 }	
		 
		 
		 //+ (padding != null ? padding.length : 0)
		 byte[] bytes = new byte[size];
		 	 
		 ByteBuffer byteBuffer = ByteBuffer.wrap(bytes);
		 
		
		 byteBuffer.put(type.byteValue());
		 byteBuffer.put((byte)semantics);
		 byteBuffer.putShort((short)(size - (padding != null ? padding.length : 0)));
		 byteBuffer.put(originatorAddress.getAddress());
		 byteBuffer.put((byte)hopLimit);
		 byteBuffer.put((byte)hopCount);
		 byteBuffer.putShort((short)sequenceNumber);


	     byteBuffer.put(messageTlvBytes);//Message TLV-block

	     
		 for (int i = 0; i < addressBlocks.size(); i++){
			 
			 byteBuffer.put(addressBytes.get(i));
			 byteBuffer.put(tlvBytes.get(i));

		 }
	     
	     if (padding != null){
	    	 byteBuffer.put(padding);
	    	 
	    	 logger.fine("Added " + padding.length + " byte(s) of padding");
	     }
	     
		 
		 return bytes;
		 
		 
	 }
}

