package net.jolsrv2.data;

import java.nio.*;
import java.util.logging.Logger;


import net.jolsrv2.constants.*;
/**
 * The class implements the TLV, as specified in OLSRv2.
 * @author Andreas Schjnhaug
 *
 */
public class Tlv {
	
	private static Logger logger = Logger.getLogger(Tlv.class.toString());
	
	/** Attributes */
	private TlvTypes type; //1 byte
	private int semantics; // 1 byte
	private int indexStart; //1 byte
	private int indexStop; //1 byte
	private byte[] value; //n bytes
	
	/**
	 * Constructor
	 * @param type the TLV type
	 */
	public Tlv(TlvTypes type){
		this.type = type;
		
	}
	
	/**
	 * Returns the TLV semantics
	 * @return  the TLV semantics
	 */
	public int getSemantics(){
		return semantics;
	}
	
	/**
	 * Sets the TLV semantics
	 * @param semantics the TLV semantics
	 */
	public void setSemantics(int semantics){
		this.semantics = semantics;
	}
	
	/**
	 * Returns the TLV value as a byte stream
	 * @return the TLV value
	 */
	public byte[] getValue(){
		return value;
	}
	
	/**
	 * Sets the TLV value as a byte stream
	 * @param value the TLV value
	 */
	public void setValue(byte[] value){
		this.value = value;
	}
	
	/**
	 * Sets the TLV value as a Value
	 * @param value the TLV value
	 */
	public void setValue(Values value){
		this.value = new byte[1];
		this.value[0] = value.byteValue();
	}
	
	/**
	 * Sets the TLV value a single byte
	 * @param value the TLV as
	 */
	public void setValue(byte value){
		this.value = new byte[1];
		this.value[0] = value;
	}	
	
	/**
	 * Returns the TLV types
	 * @return the TLV types
	 */
	public TlvTypes getType(){
		return type;
	}
	
	/**
	 * Returns the index start
	 * @return the index start
	 */
	public int getIndexStart(){
		return indexStart;
	}	
	
	
	/**
	 * Sets the index start
	 * @param indexStart the index start
	 */
	public void setIndexStart(int indexStart){
		this.indexStart = indexStart;
	}
	

	
	/**
	 * Return the index stop
	 * @return the index stop
	 */
	public int getIndexStop(){
		return indexStop;
	}
	
	/**
	 * Sets the index stop
	 * @param indexStop the index stop
	 */
	public void setIndexStop(int indexStop){
		this.indexStop = indexStop;
	}
	

	

	

	 /**
	  * Returns the bytes of the TLV
	  * @return the bytes of the TLV
	  */
	public byte[] getBytes(){
		
		logger.finest("Getting bytes from Tlv");
		logger.finest("Tlv type: " + type);
		logger.finest("Semantics: " + semantics);
		
		int size = 2; //tlvType + tlvSemantics
		
		//extended & novalue (table 1)
		//Case 1 (length is 8 bits and value is included)
		if((semantics & 0x1) == 0 && (semantics >> 1 & 0x1) == 0)
			size = size + 1 + value.length;			
		else if((semantics & 0x1) == 1 && (semantics >> 1 & 0x1) == 0)
			logger.finest("  -novalue");
		else if((semantics & 0x1) == 1 && (semantics >> 1 & 0x1) == 0){
			logger.finest("  -extended");
			size = size + 2 + value.length;
		}
		else
			logger.warning("Illegal combination of extended and novalue!");
		
		
		
		//noindex & singleindex (table 2)
		//Case 1: 
		if((semantics >> 2 & 0x1) == 0 && (semantics >> 3 & 0x1) == 0)
			size += 2;
		
		//Case 2: singleindex is set
		else if((semantics >> 2 & 0x1) == 0){
			logger.finest("  -singleindex");
			size += 1;
		}
		//Case 3: noindex is set
		else if((semantics >> 2 & 0x1) == 1 && (semantics >> 3 & 0x1) == 0){
			logger.finest("  -noindex");
		}
		else
			logger.warning("Illegal combination of noindex and singleindex");

		
		
		
		if((semantics >> 4 & 0x1) == 1){
			logger.warning("TODO: multivalue TLVs");//TODO
		}
		
		byte[] bytes = new byte[size];
		logger.finest("Bytes size: " + size);
		ByteBuffer byteBuffer = ByteBuffer.wrap(bytes);
		
		//tlv-type
		byteBuffer.put(type.byteValue());
		byteBuffer.put((byte)semantics);
		
		
		//index-start & index-stop (table 2)
		//Case 1: index-start and index-stop are included
		if((semantics >> 2 & 0x1) == 0 && (semantics >> 3 & 0x1) == 0){
			byteBuffer.put((byte)indexStart);
			byteBuffer.put((byte)indexStop);
		}
		//Case 2: index-start is included
		else if((semantics >> 2 & 0x1) == 0)
			byteBuffer.put((byte)indexStart);

		
		//length & value (table 1)
		//Case 1: length is 8 bits and value is included
		if ((semantics & 0x1) == 0 && (semantics >> 1 & 0x1) == 0){
			byteBuffer.put((byte)value.length);
			byteBuffer.put(value);
						
		}
		//Case 3: length is 16 bits and value is included
		else if ((semantics & 0x1) == 0 && (semantics >> 1 & 0x1) == 1){
			byteBuffer.putShort((short)value.length);
			byteBuffer.put(value);
		}
		
		
	     if(byteBuffer.position() != bytes.length)
	    	 logger.severe("ERROR: position != bytes.length"); //TODO Make this better and for all datatypes


		return bytes;
	}
	
	/**
	 * Checks to see if a given index is valid for an IP address in an address block
	 * @param i the index
	 * @return tue if the TLV is valid for the IP address
	 * @see AddressBlock
	 */
	public boolean isValidForAddress(int i){
		
		/**
		 * bit 2 (noindex): if cleared ('0'), the <index-start> and <index-
         stop> elements are included.  If set ('1'), the <index-start>
         and <index-stop> elements are not included.  This bit MUST be
         set for packet or message TLVs.  If this bit is set ('1') for
         address block TLVs, the TLV applies to all addresses in the
         address block.
		 */
		if((semantics >> 2 & 0x1) == 1)
			return true;
		// Checks if the i is within range
		else if(indexStart <= i && indexStop < i)
			return true;
		else		
			return false;
	}
	

}

