package net.jolsrv2.repository.neighbourhood;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.TreeSet;
import java.util.logging.Logger;

import net.jolsrv2.CidrAddress;
import net.jolsrv2.constants.TlvTypes;
import net.jolsrv2.constants.Values;
import net.jolsrv2.data.Tlv;

/**
 * The class implements the link set, as specified in OLSRv2.
 * @author Andreas Schjnhaug
 *
 */
public final class LinkSet {
	
	private static Logger logger = Logger.getLogger(LinkSet.class.toString());
	
	private static LinkSet ref;
	
	private TreeSet<LinkTuple> tuples;
	
	private LinkSet(){
		tuples = new TreeSet<LinkTuple>();
		
	}
	/**
	 * Implementation of the singleton design pattern
	 * @return the link set
	 */
	public static synchronized LinkSet getLinkSet(){
		if(ref == null)
			ref = new LinkSet();
		return ref;
	}
	
	/**
	 * Implementation of the singleton design pattern. This method is used to
	 * ensure that the class can not be cloned.
	 */
	public Object clone() throws CloneNotSupportedException{
		throw new CloneNotSupportedException();
	}
	
	/**
	 * Purges the entries in the set
	 * @return true if a change is detected
	 */
	public synchronized boolean purge(){
		boolean hasChanged = false;
		try{
			for(Iterator<LinkTuple> i = tuples.iterator(); i.hasNext(); ){
				LinkTuple link = i.next();
				if (link.getTime() < System.nanoTime()){
					logger.fine("- LinkSet: Tuple to " + link.getNeighbourCAddress() + " expired");
					i.remove();
					hasChanged = true;
					
				}
				
			}
		}
		catch(Exception e){logger.warning("ERROR in purgeLinkSet: " + e);}
		
		return hasChanged;
	}
	
	/**
	 * Populating the link set witht the addresses found in the LIB.
	 * @param receivingCAddress the receiving CIDR address
	 * @param sourceCAddress the source CIDR address
	 * @param validityTime the validity time
	 * @return true if a change in the link set is detected
	 */
	public synchronized boolean populate1(CidrAddress receivingCAddress, CidrAddress sourceCAddress, long validityTime){

		LinkTuple linkTuple = null;
	
		for(LinkTuple lt : tuples){
			if (lt.getLocalCAddress().equals(receivingCAddress) && lt.getNeighbourCAddress().equals(sourceCAddress)){
				linkTuple = lt;
				return false;
			}
		}
			
		long symTime = System.nanoTime() - (long)1e9; //Expired
		long time = System.nanoTime() + validityTime;
		linkTuple = new LinkTuple(receivingCAddress,sourceCAddress,symTime,time);
		tuples.add(linkTuple);
		
		logger.fine("- LinkSet: Added tuple from " + receivingCAddress + " to " + sourceCAddress + " (" + linkTuple.getStatus() + ")");
		
		return true;
	}
	
	/**
	 * Populating the link set with the addresses found in the address
	 * blocks other than the LIB.
	 * 
	 * @param receivingCAddress the receiving CIDR address
	 * @param sourceCAddress the source CIDR address
	 * @param validityTime the validity time
	 * @param intervalTime the interval time
	 * @param willingness the willingness
	 * @param cidrAddresses the CIDR addresses
	 * @param tlvs the TLVs
	 */
	public synchronized void populate2(CidrAddress receivingCAddress, CidrAddress sourceCAddress, long validityTime, long intervalTime, Values willingness, ArrayList<CidrAddress> cidrAddresses, ArrayList<Tlv> tlvs){	
		
		LinkTuple linkTuple = null;
		
		
		for(LinkTuple lt : tuples){
			if (lt.getLocalCAddress().equals(receivingCAddress) && lt.getNeighbourCAddress().equals(sourceCAddress)){
				linkTuple = lt;
				break;
			}
		}
		
		for(int i = 0; i < cidrAddresses.size(); i++){
			
			if(cidrAddresses.get(i).equals(receivingCAddress)){
				boolean lostFlag = false;
				boolean symmetricOrAsymmetricFlag = false;
				for(Tlv tlv : tlvs){
					if(tlv.getSemantics() == 0 && i >= tlv.getIndexStart() && i <= tlv.getIndexStop()){
						if(tlv.getType() == TlvTypes.LINK_STATUS){
							if (tlv.getValue()[0] == Values.LOST.byteValue())
								lostFlag = true;
							else if (tlv.getValue()[0] == Values.SYMMETRIC.byteValue() || tlv.getValue()[0] == Values.HEARD.byteValue())
								symmetricOrAsymmetricFlag = true;
							
						}
					}
				}
				
				if(lostFlag)
					linkTuple.setSymTime(System.nanoTime() - (long)1e9); //Expired
				else if(symmetricOrAsymmetricFlag){
					long lSymTime = System.nanoTime() + validityTime;
					linkTuple.setSymTime(lSymTime);
					linkTuple.setTime(lSymTime + intervalTime);
				}
			}
			
			
		}
		
		long lAsymTime = System.nanoTime() + validityTime;
		linkTuple.setAsymTime(lAsymTime);
		linkTuple.setTime((linkTuple.getTime()>lAsymTime ? lAsymTime : linkTuple.getTime()));
		linkTuple.setWillingness(willingness);
		
		logger.fine("- LinkSet: Updated tuple to " + sourceCAddress + " (" + linkTuple.getStatus() + ")");
	
	}
	
	/**
	 * Searches for and returns link tuples having a certain local CIDR address
	 * @param cidrBlock the local CIDR address to look for
	 * @return an ArrayList of the link tuples found
	 */
	public synchronized ArrayList<LinkTuple> getTuples(CidrAddress cidrBlock){
		
		ArrayList<LinkTuple> lts = new ArrayList<LinkTuple>(); 
		for(LinkTuple linkTuple : tuples){
			if(linkTuple.getLocalCAddress().equals(cidrBlock))
				lts.add(linkTuple);
		}
			
		
		
		return lts;
	}
	
	/**
	 * Searches for and returns link tuples having a certain TLV value
	 * @param v the TLV value to look for
	 * @return an ArrayList of the link tuples found
	 */
	public ArrayList<LinkTuple> getTuples(Values v){
		ArrayList<LinkTuple> lts = new ArrayList<LinkTuple>(); 
		for(LinkTuple linkTuple : tuples){
			if(linkTuple.getStatus().equals(v))
				lts.add(linkTuple);
		}
			
		return lts;
	}
	
	/**
	 * Returns the willigness of a link tuple with a certain local CIDR address
	 * @param from the local CIDR address
	 * @return the willingness, or null if the link tuple is not found
	 */
	public Values getWillingness(CidrAddress from){
		for(LinkTuple tuple : tuples)
			if(tuple.getLocalCAddress().equals(from))
				return tuple.getWillingness();
		
		return null;
	}
	
	/**
	 * Checks if the link set has a link tuple with a certain local CIDR address
	 * @param from the local CIDR address to look for
	 * @return true if the link tuple is found, false otherwise
	 */
	public synchronized boolean contains(CidrAddress from){
		for(LinkTuple linkTuple : tuples){
			if(linkTuple.getLocalCAddress().equals(from))
				return true;
			
		}
		return false;
	}
	
	/**
	 * Checks if the link set has a link tuple with a certain local CIDR address, neighbour CIDR address and link status
	 * @param from the local CIDR address
	 * @param to the neighbour CIDR address
	 * @param status the link status
	 * @return true if the link tuple exists, false otherwise
	 */
	public synchronized boolean contains(CidrAddress from, CidrAddress to, Values status){
		for(LinkTuple linkTuple : tuples){
			if(linkTuple.getLocalCAddress().equals(from) && linkTuple.getNeighbourCAddress().equals(to) && linkTuple.getStatus().equals(status))
				return true;
			
		}
		return false;
	}
	
	/**
	 * Returns the link tuples in the link set
	 * @return the link tuples
	 */
	public TreeSet<LinkTuple> getTuples(){
		return tuples;
	}
	
}
