package net.jolsrv2.repository.neighbourhood;

import java.util.Comparator;
import java.util.logging.Logger;

import net.jolsrv2.CidrAddress;
import net.jolsrv2.constants.*;
/**
 * The class implements the link tuple, as specified in OLSRv2.
 * @author Andreas Schjnhaug
 *
 */
public class LinkTuple implements Comparator<LinkTuple>{
	
	private static Logger logger = Logger.getLogger(LinkTuple.class.toString());
	
    
	private CidrAddress localCAddress;
	private CidrAddress neighbourCAddress;
	private long symTime;
	private long asymTime;
	private Values willingness;
	private long time;
	
	/**
	 * Constructor
	 * @param localCAddress the local CIDR address
	 * @param neighbourCAddress the neighbour CIDR address
	 * @param symTime the symmetric time
	 * @param time the time
	 */
	public LinkTuple(CidrAddress localCAddress, CidrAddress neighbourCAddress, long symTime, long time){
		this.localCAddress = localCAddress;
		this.neighbourCAddress = neighbourCAddress;
		this.symTime = symTime;
		this.time = time;
		
	}
	
	/**
	 * Returns the neighbour CIDR address
	 * @return the neighbour CIDR address
	 */
	public CidrAddress getNeighbourCAddress(){
		return neighbourCAddress;
		
	}
	
	/**
	 * Returns the local CIDR address
	 * @return the local CIDR address
	 */
	public CidrAddress getLocalCAddress(){
		return localCAddress;
		
	}
	
	/**
	 * Return the time
	 * @return the time
	 */
	public long getTime(){
		return time;
	}
	
	/**
	 * Sets the time
	 * @param time the time
	 */
	public void setTime(long time){
		this.time = time;
	}
	

	/**
	 * Return the willingness
	 * @return the willingness
	 */
	public Values getWillingness(){
		return willingness;
	}
	
	/**
	 * Sets the willingness
	 * @param willingness the willingness
	 */
	public void setWillingness(Values willingness){
		this.willingness = willingness;
	}
	
	/**
	 * Gets the symmetric time
	 * @return the symmetric time
	 */
	public long getSymTime(){
		return symTime;
	}
	
	/**
	 * Sets the symmetric time
	 * @param symTime the symmetric time
	 */
	public void setSymTime(long symTime){
		this.symTime = symTime;
	}	
	
	/**
	 * Gets the aymmetric time
	 * @return the asymmetric time
	 */	
	public long getAsymTime(){
		return asymTime;
	}
	
	/**
	 * Sets the asymmetric time
	 * @param asymTime the asymmetric time
	 */	
	public void setAsymTime(long asymTime){
		this.asymTime = asymTime;
	}
	
	/**
	 * Gets the link status
	 * @return the link status
	 */
	public Values getStatus(){
		if(symTime < System.nanoTime() && asymTime < System.nanoTime()){
			return Values.LOST;			
		}
		else if (symTime > System.nanoTime()){
			return Values.SYMMETRIC;						
		}
		else if (symTime < System.nanoTime() && asymTime > System.nanoTime()){
			return Values.HEARD;							
		}
		logger.warning("ERROR in LinkTuple: Neither LOST, SYMMETRIC nor ASSYMETRIC");
		return null;
	}
	
	/**
	 * Compares the components of two link tuples, to determine how they should be sorted
	 * @param l1 the first link tuple to compare
	 * @param l2 the first link tuple to compare
	 * @return -1 if l1 should be sorted before l2, 1 if l2 should be sorted before l1, or 0 if they are equal
	 */
	public int compare(LinkTuple l1, LinkTuple l2){
		if(l1.getNeighbourCAddress().getPrefixLength() < l2.getNeighbourCAddress().getPrefixLength())
			return -1;
		else if(l1.getNeighbourCAddress().getPrefixLength() > l2.getNeighbourCAddress().getPrefixLength())
			return 1;
		return 0;
	}
	
	/**
	 * Compares the components of two link tuples, to determine if they are equal
	 * @param tuple the link tuple to compare with this link tuple
	 * @return true if all the components of the two link tuples are equal
	 */
	public boolean equals(LinkTuple tuple){
		return(localCAddress.equals(tuple.getLocalCAddress()) && neighbourCAddress.equals(tuple.getNeighbourCAddress()) && symTime == tuple.getSymTime() && asymTime == tuple.getAsymTime() && willingness.equals(tuple.getWillingness()) && time == tuple.getTime());
	}
}

