package net.jolsrv2.repository.neighbourhood;


import java.util.ArrayList;
import java.util.Iterator;
import java.util.logging.Logger;

import net.jolsrv2.CidrAddress;
import net.jolsrv2.Interface;
import net.jolsrv2.Node;
import net.jolsrv2.constants.*;
import net.jolsrv2.data.Tlv;

/**
 * The class implements the MPR selector set, as specified in OLSRv2.
 * @author Andreas Schjnhaug
 *
 */

public final class MprSelectorSet {
	
	private static Logger logger = Logger.getLogger(MprSelectorSet.class.toString());
	
	private static MprSelectorSet ref;
	
	
	private ArrayList<MprSelectorTuple> tuples;
	
	private MprSelectorSet(){
		tuples = new ArrayList<MprSelectorTuple>();
	}
	
	/**
	 * Implementation of the singleton design pattern
	 * @return the MPR selector set
	 */
	public static synchronized MprSelectorSet getMprSelectorSet(){
		if(ref == null)
			ref = new MprSelectorSet();
		return ref;
	}
	
	/**
	 * Implementation of the singleton design pattern. This method is used to
	 * ensure that the class can not be cloned.
	 */	
	public Object clone() throws CloneNotSupportedException{
		throw new CloneNotSupportedException();
	}
	
	/**
	 * Populates the MPR selector set.
	 * @param localCAddresses the local CIDR address
	 * @param libCAddresses the LIB CIDR addresses
	 * @param cidrAddresses the CIDR addresses
	 * @param tlvs the TLVs
	 * @param validityTime the validity time
	 * @return true if a change in the MPR selector set is detected
	 */
	public synchronized boolean populate(CidrAddress localCAddresses, ArrayList<CidrAddress> libCAddresses, ArrayList<CidrAddress> cidrAddresses, ArrayList<Tlv> tlvs, long validityTime){
		
		boolean hasChanged = false;
		
		ArrayList<CidrAddress> interfaceCAddresses = new ArrayList<CidrAddress>();
		for(Interface iface : Node.getNode().getInterfaces())
			interfaceCAddresses.add(iface.getLocalCAddress());
		

		for(Tlv tlv : tlvs){
			if(tlv.getType().equals(TlvTypes.MPR)){
				
				for(int i = 0; i < cidrAddresses.size(); i++){
					if(cidrAddresses.get(i).equals(localCAddresses) && tlv.isValidForAddress(i)){
						
						//8.3 All LIB CIDR must be updated or added
						for(CidrAddress libAddress : libCAddresses){
							
							MprSelectorTuple found = null;
							for(MprSelectorTuple relayTuple : tuples){
								if(relayTuple.getCAddress().equals(libAddress)){
									found = relayTuple;
									//TODO: break??
								}
							}
							/*Sout.print("-------------------------------");
							Sout.print("  Populating MPR Selector set  ");
							Sout.print("-------------------------------");*/
							if(found == null){
								found = new MprSelectorTuple(libAddress, System.nanoTime() + validityTime);
								tuples.add(found);
								hasChanged = true;
								logger.fine("        - MprSelectorSet: Added tuple to " + libAddress);
							}
							else{
								found.setTime(System.nanoTime() + validityTime);
								logger.fine("        - MprSelectorSet: Updated tuple to " + found.getCAddress());
							}
						}
							break;
					}
				}
			}
			else if(tlv.getType().equals(TlvTypes.LINK_STATUS) && tlv.getValue()[0] == Values.SYMMETRIC.byteValue()){
				for(int i = 0; i < cidrAddresses.size(); i++){
					if(cidrAddresses.get(i).equals(localCAddresses) && tlv.isValidForAddress(i)){
						try{
							for(Iterator<MprSelectorTuple> j = tuples.iterator(); j.hasNext(); ){
								MprSelectorTuple tuple = j.next();
								if (tuple.getTime() < System.nanoTime()){
									logger.fine("- MprSelectorSet: Removed tuple to " + tuple.getCAddress());
									j.remove();
									hasChanged = true;
									break;
									
								}
								
							}
						}
						catch(Exception e){logger.warning("ERROR in MprSelectorSet.populate: " + e);}
						
					}
					
				}
				
				
			}
		}
		
		return hasChanged;
	}
	

	/**
	 * Returns the CIDR addresses found in the MPR selector tuples.
	 * @return an ArrayList of the CIDR addresses
	 */
	public synchronized ArrayList<CidrAddress> getCAddresses(){
		ArrayList<CidrAddress> cidrAddresses = new ArrayList<CidrAddress>();
		for(MprSelectorTuple tuple : tuples){
			cidrAddresses.add(tuple.getCAddress());
			
		}
		return cidrAddresses;
	}
	
	/**
	 * Checks to see if a CIDR addresses is present in the MPR selector set
	 * @param cidrAddress the CIDR address
	 * @return true if found
	 */
	public synchronized boolean contains(CidrAddress cidrAddress){
		for(MprSelectorTuple tuple : tuples){
			if (tuple.getCAddress().equals(cidrAddress))
				return true;
		}
		return false;
		
	}
	
	
	/**
	 * The method purges the MPR selector set for expired entries
	 * @return true if a change is found
	 */
	public synchronized boolean purge(){
		boolean hasChanged = false;
		try{
			for(Iterator<MprSelectorTuple> i = tuples.iterator(); i.hasNext(); ){
				MprSelectorTuple tuple = i.next();
				if (tuple.getTime() < System.nanoTime()){
					logger.fine("- MprSelectorSet: Removed tuple to " + tuple.getCAddress());
					i.remove();
					hasChanged = true;
					break;
					
				}
				
			}
		}
		catch(Exception e){logger.warning("ERROR in MprSelectorSet: " + e);}
		
		return hasChanged;
	}

}
