package net.jolsrv2.repository.neighbourhood;

import java.net.InetAddress;
import java.util.*;
import java.util.logging.Logger;

import net.jolsrv2.CidrAddress;
import net.jolsrv2.constants.*;
/**
 * The class implements the MPR set, as specified in OLSRv2.
 * @author Andreas Schjnhaug
 *
 */
public class MprSet {
	
	private static Logger logger = Logger.getLogger(MprSet.class.toString());
	
	private CidrAddress ifaceCAddress;
	
	private ArrayList<InetAddress> addresses;

	/**
	 * Constructor 
	 * @param ifaceCAddress the interface CIDR address
	 */
	public MprSet(CidrAddress ifaceCAddress){
	  
		this.ifaceCAddress = ifaceCAddress;
		addresses = new ArrayList<InetAddress>();
	}
  
	
	/**
	 * Populates the MPR set, using the heuristics proposed in the specification.
	 *
	 */
	public void populate(){
		
		logger.finer("Starting populating the MPR set");
		addresses.clear();

		ArrayList<SymmetricNeighbourTuple> n = SymmetricNeighbourSet.getSymmetricNeighbourSet().getTuples(ifaceCAddress);

		//Removing neighbours with willingess == WILL_NEVER
		for(Iterator<SymmetricNeighbourTuple> i = n.iterator(); i.hasNext();){
			SymmetricNeighbourTuple nTuple = i.next();

			
			if(LinkSet.getLinkSet().contains(nTuple.getLocalCAddress(),nTuple.getNeighbourCAddress(),Values.WILL_NEVER)){

					i.remove();
					break;
			}
		}
		
		ArrayList<TwoHopTuple> n2 = TwoHopSet.getTwoHopSet().getTuples(ifaceCAddress);
		logger.finer("orgSIZE N2: " + n2.size());
		//Removing interfaces in n2 which are already in n
		for(Iterator<TwoHopTuple> i = n2.iterator(); i.hasNext();){
			TwoHopTuple t2 = i.next();
			for(SymmetricNeighbourTuple t1 : n){
				if(t1.getNeighbourCAddress().equals(t2.getTwoHopCAddress()))
					i.remove();
					
					
			}
		}
		
		
		logger.finer("SIZE N: " + n.size());
		logger.finer("SIZE N2: " + n2.size());
		
		if(n2.size() == 0)
			return;
		
		
		//Adding all members of N with willingness WILL_ALWAYS
		for(SymmetricNeighbourTuple nTuple : n){
			if(LinkSet.getLinkSet().getWillingness(nTuple.getLocalCAddress()).equals(Values.WILL_ALWAYS)){
			//if(linkTuple.getWillingness().equals(Values.WILL_ALWAYS)){
				addresses.add(nTuple.getNeighbourCAddress().getAddress());
				logger.finer("- Added MPR with willingness WILL_ALWAYS");
			}
		}
		  
		//Calculating D(y)
		HashMap<CidrAddress, Integer> dy = calculateReachability(n,n2);
		
		//Adding the links in N which are the only connection to a node in N2
		for(TwoHopTuple n2Tuple : n2){
			int counter = 0;
			for(SymmetricNeighbourTuple nTuple : n){
				if (nTuple.getNeighbourCAddress().equals(n2Tuple.getNeighbourCAddress()))
					counter++;
			}
			if (counter == 1){
				if(!addresses.contains(n2Tuple.getNeighbourCAddress())){
					addresses.add(n2Tuple.getNeighbourCAddress().getAddress());
					logger.finer("- Added MPR which is the only connection to a two hop link");
				}
			}
		}
		
		//Reducing n2
		n2 = reduceN2(n2);
		
		while(!n2.isEmpty()){
			//Calculating the reachability
			
			HashMap<CidrAddress, Integer> reachability = calculateReachability(n,n2);
			
			SymmetricNeighbourTuple candidate = null;
			Values candidateWillingness = null;
			for(SymmetricNeighbourTuple nTuple : n){
				
				if(reachability.get(nTuple.getLocalCAddress()).intValue() > 0){
					if(candidate == null){
						candidate = nTuple;
						candidateWillingness = LinkSet.getLinkSet().getWillingness(nTuple.getLocalCAddress());
					}
					else{
						
						Values willingness = LinkSet.getLinkSet().getWillingness(nTuple.getLocalCAddress());
						if(willingness.byteValue() > candidateWillingness.byteValue())
							candidate = nTuple;
						else if(willingness.byteValue() == candidateWillingness.byteValue()){
							if(reachability.get(nTuple.getLocalCAddress()) > reachability.get(candidate.getLocalCAddress()))
								candidate = nTuple;
							else if(reachability.get(nTuple.getLocalCAddress()) == reachability.get(candidate.getLocalCAddress())){
								if(dy.get(nTuple.getLocalCAddress()) > dy.get(candidate.getLocalCAddress()))
									candidate = nTuple;
							}
						}
					}
				}
			}
			if(candidate == null)
				logger.warning("ERROR in MprSet. Candidate NULL");
			addresses.add(candidate.getLocalCAddress().getAddress());
			
			//Reducing n2
			n2 = reduceN2(n2);
		
		}
		logger.finer("- The following links were chosen as MPRs:");
		for(InetAddress ia : addresses){
			logger.finer("       " + ia.toString());
		}
	}
  
	
	/**
	 * Helper method used in the calculation of the MPR set.
	 * @param n the symmetric neighbour tuples
	 * @param n2 the two hop tuples
	 * @return A HashMap showing the reachability
	 */
	private HashMap<CidrAddress, Integer> calculateReachability(ArrayList<SymmetricNeighbourTuple> n, ArrayList<TwoHopTuple> n2){
		HashMap<CidrAddress, Integer> reachability = new HashMap<CidrAddress, Integer>();
		for(SymmetricNeighbourTuple linkTuple : n){
			int counter = 0;
			for(TwoHopTuple twoHopTuple : n2)
				if (linkTuple.getNeighbourCAddress().equals(twoHopTuple.getNeighbourCAddress()))
					counter++;
			reachability.put(linkTuple.getLocalCAddress(), new Integer(counter));
		}
		
		return reachability;
	  
	}
	
	/**
	 * Helper method used in the calculation of the MPR set.
	 * @param n2 the two hop tuples
	 * @return An ArrayList containing the reduced n2 set
	 */
	private ArrayList<TwoHopTuple> reduceN2(ArrayList<TwoHopTuple> n2){
		
		//Removing the interfaces from N2 which are covered by an interface in the MPR set
		try{
			for(Iterator<TwoHopTuple> i = n2.iterator(); i.hasNext();){
				TwoHopTuple twoHopTuple = i.next();
				for(InetAddress address : addresses){
					if(address.equals(twoHopTuple.getNeighbourCAddress())){
						i.remove();
						break;
					}
				}				  
			}
		}
		catch(Exception e){logger.warning("ERROR in MprSet: " + e);}
		
		return n2;
		
	}
	
	/**
	 * Returns the IP addresses
	 * @return the IP addresses
	 */
	public ArrayList<InetAddress> getAddresses(){
		return addresses;
	}
	
	/**
	 * Determines whether the MPR set is empty or not
	 * @return true if the MPR set is empty
	 */
	public boolean isEmpty(){
		return addresses.isEmpty();
	}

}

