package net.jolsrv2.repository.neighbourhood;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.logging.Logger;

import net.jolsrv2.CidrAddress;
/**
 * The class implements the neighbour address association set, as specified in OLSRv2.
 * @author Andreas Schjnhaug
 *
 */
public final class NaaSet {
	
	private static Logger logger = Logger.getLogger(NaaSet.class.toString());
	
	private static NaaSet ref;
	
	private ArrayList<NaaTuple> tuples;

	/**
	 * Implementation of the singleton design pattern
	 * @return the neighbour address association set
	 */
	public static synchronized NaaSet getNaaSet(){
		if(ref == null)
			ref = new NaaSet();
		return ref;
	}
	/**
	 * Implementation of the singleton design pattern. This method is used to
	 * ensure that the class can not be cloned.
	 */
	public Object clone() throws CloneNotSupportedException{
		throw new CloneNotSupportedException();
	}
	
	/**
	 * Constructor
	 *
	 */
	private NaaSet(){
		tuples = new ArrayList<NaaTuple>();
	}
	
	/**
	 * Method for populating the neighbour address association set
	 * @param intervalTime the interval time
	 * @param cidrAddresses the CIDR addresses
	 * @return true if a change is detected
	 */
	public boolean populate(long intervalTime, ArrayList<CidrAddress> cidrAddresses){
		
		
		//TODO How should this be reflected in updating the routing set

		boolean change = false;
		
	

		//Removing old entries
		try{
			for(Iterator<NaaTuple> i = tuples.iterator(); i.hasNext(); ){
				NaaTuple tuple = i.next();
				
				boolean found = false;
				for(CidrAddress a1 : tuple.getCAddresses()){
					for(CidrAddress a2 : cidrAddresses){
						if(a1.equals(a2))
							found = true;
					}
				}
				if(found){
					i.remove();
					change = true;
				}
			}
		}
		catch(Exception e){logger.warning("ERROR in NaaSet: " + e);}
		
		//Adding tuple
		tuples.add(new NaaTuple(cidrAddresses, System.nanoTime() + 3*intervalTime)); //TODO: shoudl be NA_HOLD_TIME = 2 x TC_INTERVAL 
		logger.fine("- NAA set: Added tuple to " + cidrAddresses);
		
		return change;

	}
	/**
	 * The method purges the neighbour address association set, deleting expired entries
	 *
	 */
	public void purge(){
		try{
			for(Iterator<NaaTuple> i = tuples.iterator(); i.hasNext(); ){
				NaaTuple tuple = i.next();
				if (tuple.getTime() < System.nanoTime()){
					logger.fine("- NaaSet: Removed tuple to " + tuple.getCAddresses());
					i.remove();
					
				}
				
			}
		}
		catch(Exception e){logger.warning("ERROR in NeighbourhoodAddressAssociationTuple.purge(): " + e);}
	}
	
	/**
	 * Returns the neighbour address association tuples
	 * @return the neighbour address association tuples
	 */
	public ArrayList<NaaTuple> getTuples(){
		return tuples;
	}
	
	

}
