package net.jolsrv2.repository.neighbourhood;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.logging.Logger;

import net.jolsrv2.CidrAddress;
import net.jolsrv2.constants.TlvTypes;
import net.jolsrv2.constants.Values;
import net.jolsrv2.data.Tlv;

/**
 * The class implements the symmetric neighbour set, as specified in OLSRv2.
 * @author Andreas Schjnhaug
 *
 */

public class SymmetricNeighbourSet {
	
	private static Logger logger = Logger.getLogger(SymmetricNeighbourSet.class.toString());
	
	private static SymmetricNeighbourSet ref;
	
	private ArrayList<SymmetricNeighbourTuple> tuples;
	
	/**
	 * Constructor
	 *
	 */
	public SymmetricNeighbourSet(){
		tuples = new ArrayList<SymmetricNeighbourTuple>();
	}
	
	/**
	 * Implementation of the singleton design pattern.
	 * @return the symmetric neighbour set
	 */
	public static synchronized SymmetricNeighbourSet getSymmetricNeighbourSet(){
		if(ref == null)
			ref = new SymmetricNeighbourSet();
		return ref;
	}
	
	
	/**
	 * Implementation of the singleton design pattern. This method is used to
	 * ensure that the class can not be cloned.
	 */
	public Object clone() throws CloneNotSupportedException{
		throw new CloneNotSupportedException();
	}
	
	/**
	 * Method populating the symmetric neighbour set, as proposed in the specification
	 * @param receivingCAddress the receiving CIDR address
	 * @param libCAddresses the LIB CIDR address
	 * @param cidrAddresses the CIDR addresses
	 * @param tlvs the TLVs
	 * @param validityTime the validity time
	 */
	public void populate(CidrAddress receivingCAddress, ArrayList<CidrAddress> libCAddresses, ArrayList<CidrAddress> cidrAddresses, ArrayList<Tlv> tlvs, long validityTime){
		//6.2 in ND
		for(CidrAddress cidrAddress : cidrAddresses){
			if(cidrAddress.equals(receivingCAddress))
				for(Tlv tlv : tlvs){
					if(tlv.getType().equals(TlvTypes.LINK_STATUS)){
						if (tlv.getValue()[0] == Values.SYMMETRIC.byteValue() || tlv.getValue()[0] == Values.HEARD.byteValue()){
							for(CidrAddress libAddress : libCAddresses){
								SymmetricNeighbourTuple tuple = getTuple(receivingCAddress,libAddress);
								
								if(tuple != null){
									
									tuple.setSymTime(System.nanoTime() + validityTime);
									tuple.setTime(tuple.getSymTime() + (long)(6*1e9)); //TODO N_time = N_SYM_time + N_HOLD_TIME.
									
								}
								else{
									SymmetricNeighbourTuple newTuple = new SymmetricNeighbourTuple(receivingCAddress, libAddress, System.nanoTime() + validityTime, System.nanoTime() + (long)(6*1e9) ); //TODO same as above
									tuples.add(newTuple);
								}
								
							}
							
						}
						else if(tlv.getValue()[0] == Values.LOST.byteValue()){
							for(CidrAddress libAddress : libCAddresses){
								SymmetricNeighbourTuple tuple = getTuple(receivingCAddress,libAddress);
								
								if (tuple != null){
									tuple.setSymTime(System.nanoTime() - (long)1e9); //Expired
									
									tuple.setTime(Math.min(tuple.getTime(), System.nanoTime() + (long)(6*1e9))); //TODO same as above
								}		
							}
						}	
					}
				}	
		}
		
	}
	
	/**
	 * Returns the tuple with a certain local CIDR address and neighbour CIDR address
	 * @param from the local CIDR address
	 * @param to the neighbour CIDR address
	 * @return a symmetric neighbour tuple if found, or null if not
	 */
	private SymmetricNeighbourTuple getTuple(CidrAddress from, CidrAddress to){
		
		
		for(SymmetricNeighbourTuple tuple : tuples)
			if(tuple.getLocalCAddress().equals(from) && tuple.getNeighbourCAddress().equals(to))
					return tuple;
			
		
		
		return null;
		
	}
	
	/**
	 * Returns the tuples with a certain local CIDR address
	 * @param from the local CIDR address
	 * @return an ArrayList containing the symmetric neighbour tuples matching the criterias
	 */
	public ArrayList<SymmetricNeighbourTuple> getTuples(CidrAddress from){
		ArrayList<SymmetricNeighbourTuple> returnTuples = new ArrayList<SymmetricNeighbourTuple>();
		for(SymmetricNeighbourTuple tuple : tuples)
			if(tuple.getLocalCAddress().equals(from))
				returnTuples.add(tuple);
		
		
		
		return returnTuples;
	}
	
	/**
	 * The method purges the symmetric neighbour set, deleting expired entries
	 * @return true if a change is detected
	 */
	public synchronized boolean purge(){
		boolean hasChanged = false;
		try{
			for(Iterator<SymmetricNeighbourTuple> i = tuples.iterator(); i.hasNext(); ){
				SymmetricNeighbourTuple tuple = i.next();
				logger.finer("NANOTIME: " + System.nanoTime());
				logger.finer("tupletim: " + tuple.getTime());
				if (tuple.getTime() < System.nanoTime()){
					logger.fine("- SymmetricNeighbourSet: Tuple from " + tuple.getLocalCAddress() + " to " + tuple.getNeighbourCAddress() + " expired");
					i.remove();
					hasChanged = true;
					
				}
				
			}
		}
		catch(Exception e){logger.warning("ERROR in SymmetricNeighbourSet.purge(): " + e);}
		
		return hasChanged;
	}

}
