package net.jolsrv2.repository.neighbourhood;

import java.util.*;
import java.util.logging.Logger;

import net.jolsrv2.CidrAddress;
import net.jolsrv2.constants.*;
import net.jolsrv2.data.Tlv;


/**
 * The class implements the two hop set, as specified in OLSRv2.
 * @author Andreas Schjnhaug
 *
 */

public final class TwoHopSet {
	private static Logger logger = Logger.getLogger(TwoHopSet.class.toString());
	
	
	private static TwoHopSet ref;
	
	private ArrayList<TwoHopTuple> tuples;
	
	
	/**
	 * Constructor
	 *
	 */
    private TwoHopSet(){
    	tuples = new ArrayList<TwoHopTuple>();
    	
    }
    
    /**
	 * Implementation of the singleton design pattern
	 * @return the two hop set
	 */
    public static synchronized TwoHopSet getTwoHopSet(){
		if(ref == null)
			ref = new TwoHopSet();
		return ref;
	}
	
    /**
	 * Implementation of the singleton design pattern. This method is used to
	 * ensure that the class can not be cloned.
	 */
	public Object clone() throws CloneNotSupportedException{
		throw new CloneNotSupportedException();
	}
    
	
	
	/**
	 * Populates the two hop set as suggested in the specification
	 * @param receivingCAddress the receiving CIDR address
	 * @param sendingCAddress the sending CIDR address
	 * @param validityTime the validity time
	 * @param cidrAddresses the CIDR address
	 * @param tlvs the TLVs
	 * @return true if a change in the two hop set is detected
	 */
    public boolean populate(CidrAddress receivingCAddress, CidrAddress sendingCAddress, long validityTime, ArrayList<CidrAddress> cidrAddresses, ArrayList<Tlv> tlvs){
    
    
    	boolean hasChanged = false;
    	
    	//Check to see if the message is from a symmetric neighbour
    	if(LinkSet.getLinkSet().contains(receivingCAddress, sendingCAddress, Values.SYMMETRIC)){
    		for(Tlv tlv : tlvs){
    			
    			
    			if(tlv.getType() == TlvTypes.LINK_STATUS && tlv.getValue()[0] == Values.SYMMETRIC.byteValue()){
					for(int j = tlv.getIndexStart(); j <= tlv.getIndexStop(); j++){
						if(!cidrAddresses.get(j).equals(receivingCAddress)){
							TwoHopTuple twoHopTuple = getTuple(receivingCAddress, sendingCAddress, cidrAddresses.get(j));
							if(twoHopTuple == null){
								twoHopTuple = new TwoHopTuple(receivingCAddress, sendingCAddress, cidrAddresses.get(j), System.nanoTime() + validityTime);
								tuples.add(twoHopTuple);
								hasChanged = true;
								logger.fine("- TwoHopSet: Added tuple to " + twoHopTuple.getTwoHopCAddress());
							}
							else{
								twoHopTuple.setTime(System.nanoTime() + validityTime);
								logger.fine("- TwoHopSet: Updated tuple to " + twoHopTuple.getTwoHopCAddress());
							}
						}
					}
				}
    		}
    	}
		return hasChanged;
    }
    /**
     * The method gets all two hop tuples associated with a certain local CIDR address
     * @param cidrBlock the local CIDR address
     * @return an ArrayList of two hop tuples associated with an address
     */
    public ArrayList<TwoHopTuple> getTuples(CidrAddress cidrBlock){
    	ArrayList<TwoHopTuple> foundTuples = new ArrayList<TwoHopTuple>(); 
    	for(TwoHopTuple twoHopTuple : tuples){
    		if(twoHopTuple.getLocalCAddress().equals(cidrBlock))
    			foundTuples.add(twoHopTuple);
    	}
    	return foundTuples;
    	
    }
    

    /**
     * The method gets all two hop tuples associated with a certain local CIDR address, neighbourhood CIDR address and two hop CIDR address
     * @param i1 the local CIDR address
     * @param i2 the neighbourhood CIDR address
     * @param i3 the two hop CIDR address
     * @return the two hop tuple if found, null if not
     */
    private TwoHopTuple getTuple(CidrAddress i1, CidrAddress i2, CidrAddress i3){
    	for(TwoHopTuple twoHopTuple : tuples){
    		if(twoHopTuple.getLocalCAddress().equals(i1) && twoHopTuple.getNeighbourCAddress().equals(i2) && twoHopTuple.getTwoHopCAddress().equals(i3))
    			return twoHopTuple;
    	}
    	return null;
    	
    }
    
	/**
	 * The method purges the set, deleting expired entries
	 * @return true if a change is detected
	 */
    public boolean purge(){
		boolean hasChanged = false;

		try{
			for(Iterator<TwoHopTuple> i = tuples.iterator(); i.hasNext(); ){
				TwoHopTuple tuple = i.next();
				if (tuple.getTime() < System.nanoTime()){
					logger.fine("- TwoHopSet: Removed tuple to " + tuple.getNeighbourCAddress());
					i.remove();
					hasChanged = true;
					
				}
				
			}
		}
		catch(Exception e){logger.warning("ERROR in purgeLinkSet: " + e);}
		
		return hasChanged;
	}
	
    /**
     * Returns the two hop tuples
     * @return the two hop tuples
     */
    public ArrayList<TwoHopTuple> getTuples(){
    	return tuples;
    }
    
    
    


}

