package net.jolsrv2.repository.topology;

import java.net.InetAddress;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.logging.Logger;
/**
 * The class implements the ANSN history set, as specified in OLSRv2.
 * @author Andreas Schjnhaug
 *
 */
public class AnsnHistorySet {
	
	private static Logger logger = Logger.getLogger(AnsnHistorySet.class.toString());
	
	private ArrayList<AnsnHistoryTuple> tuples;
	
	private static AnsnHistorySet ref;
	
	/**
	 * Constructor
	 *
	 */
	public AnsnHistorySet(){
		tuples = new ArrayList<AnsnHistoryTuple>();
	}
	
	
	/**
	 * Implementation of the singleton design pattern
	 * @return the ANSN history set
	 */
	public static synchronized AnsnHistorySet getAnsnHistorySet(){
		if(ref == null)
			ref = new AnsnHistorySet();
		return ref;
	}
	
	/**
	 * Implementation of the singleton design pattern. This method is used to
	 * ensure that the class can not be cloned.
	 */
	public Object clone() throws CloneNotSupportedException{
		throw new CloneNotSupportedException();
	}
	
	/**
	 * Method populating the advertised neighbour set, as proposed in the specification
	 */
	public boolean populate(InetAddress originatorAddress, int ansn, long validityTime){
		for(AnsnHistoryTuple tuple : tuples){
			if(tuple.getOriginatorAddress().equals(originatorAddress)){
				if(tuple.getAnsn() > ansn)
					return false;
				else{
					tuple.setTime(System.nanoTime() + validityTime);
					tuple.setAnsn(ansn);
					return true;
				}
				
			}
		}
		tuples.add(new AnsnHistoryTuple(originatorAddress,ansn, System.nanoTime() + validityTime));
		return true;
		
	}
	/**
	 * The method purges the ANSN history set, deleting entries with expired time
	 * 
	 */
	public void purgeTime(){
		try{
			for(Iterator<AnsnHistoryTuple> i = tuples.iterator(); i.hasNext(); ){
				AnsnHistoryTuple tuple = i.next();
				if (tuple.getTime() < System.nanoTime()){
					logger.fine("- AnsnHistorySet: Tuple to " + tuple.getOriginatorAddress() + " expired");
					i.remove();
				}
			}
		}
		catch(Exception e){logger.warning("ERROR in AnsnHistorySet.purge(): " + e);}
		
	}
	/**
	 * The method purges the ANSN history set, deleting entries with expired ANSN
	 * 
	 */
	public void purgeAnsn(){
		try{
			for(Iterator<AnsnHistoryTuple> i = tuples.iterator(); i.hasNext(); ){
				AnsnHistoryTuple tuple = i.next();
				if (tuple.getTime() < System.nanoTime()){
					logger.fine("- AnsnHistorySet: Tuple to " + tuple.getOriginatorAddress() + " expired");
					i.remove();
				}
			}
		}
		catch(Exception e){logger.warning("ERROR in AnsnHistorySet.purge(): " + e);}
		
	}

}
