package net.jolsrv2.repository.topology;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.logging.Logger;

import net.jolsrv2.CidrAddress;
import net.jolsrv2.constants.TlvTypes;
import net.jolsrv2.data.Tlv;

/**
 * The class implements the attached network set, as specified in OLSRv2.
 * @author Andreas Schjnhaug
 *
 */
public class AttachedNetworkSet {
	
	private static Logger logger = Logger.getLogger(AttachedNetworkSet.class.toString());
	
	
	private ArrayList<AttachedNetworkTuple> tuples;
	private static AttachedNetworkSet ref;
	
	/**
	 * Constructor
	 *
	 */
	public AttachedNetworkSet(){
		tuples = new ArrayList<AttachedNetworkTuple>();
	}
	
	/**
	 * Implementation of the singleton design pattern
	 * @return the attached network set
	 */
	public static synchronized AttachedNetworkSet getAttachedNetworkSet(){
		if(ref == null)
			ref = new AttachedNetworkSet();
		return ref;
	}
	
	/**
	 * Implementation of the singleton design pattern. This method is used to
	 * ensure that the class can not be cloned.
	 */
	public Object clone() throws CloneNotSupportedException{
		throw new CloneNotSupportedException();
	}
	/**
	 * Method populating the symmetric neighbour set, as proposed in the specification
	 * @param libCAddresses the LIB CIDR addresses
	 * @param cAddresses the CIDR addresses
	 * @param tlvs the TLVs
	 * @param ansn the ANSN
	 * @param validityTime the validity time
	 */
	public void populate(ArrayList<CidrAddress> libCAddresses, ArrayList<CidrAddress> cAddresses, ArrayList<Tlv> tlvs, int ansn, long validityTime){
		for(Tlv tlv : tlvs){
			if(tlv.getType().equals(TlvTypes.GATEWAY)){
				for(int i = 0; i < cAddresses.size(); i++){
					if(tlv.isValidForAddress(i)){
						for(CidrAddress libCAddress : libCAddresses){
							AttachedNetworkTuple tuple = getTuple(cAddresses.get(i), libCAddress);
							if(tuple != null){
								tuple.setAnsn(ansn);
								tuple.setTime(System.nanoTime() + validityTime);
							}
							else{
								tuples.add(new AttachedNetworkTuple(cAddresses.get(i), libCAddress,ansn, System.nanoTime() + validityTime));
								logger.fine("- AttachedNetworkSet: Added tuple to network " + cAddresses.get(i) + " through gateway " + libCAddress);
								
							}
						}
					}
				}
			}
		}
	}
	
	/**
	 * Returns an attached network tuple containing a certain network CIDR address and gateway CIDR address
	 * @param networkCAddress the network CIDR address
	 * @param gatewayCAddress the gateway CIDR address
	 * @return the attached network tuple if it exists
	 */
	private AttachedNetworkTuple getTuple(CidrAddress networkCAddress, CidrAddress gatewayCAddress){
		for(AttachedNetworkTuple tuple : tuples){
			if(tuple.getNetworkCAddress().equals(networkCAddress) && tuple.getGatewayCAddress().equals(gatewayCAddress)){
				return tuple;
			}
		}
		return null;
	}
	
	/**
	 * The method purges the attached network set, deleting expired entries
	 */
	public void purge(){
		try{
			for(Iterator<AttachedNetworkTuple> i = tuples.iterator(); i.hasNext(); ){
				AttachedNetworkTuple tuple = i.next();
				if (tuple.getTime() < System.nanoTime()){
					logger.fine("- AttachedNetworkSet: Tuple to network " + tuple.getNetworkCAddress() + " through gateway " + tuple.getGatewayCAddress() + " removed");
					i.remove();
				}
			}
		}
		catch(Exception e){logger.warning("ERROR in AttachedNetworkSet.purge(): " + e);}
		
	}	
}
