package net.jolsrv2.repository.topology;

import java.net.InetAddress;
import java.util.*;
import java.util.logging.Logger;

import net.jolsrv2.CidrAddress;
import net.jolsrv2.constants.Values;
import net.jolsrv2.repository.neighbourhood.LinkSet;
import net.jolsrv2.repository.neighbourhood.LinkTuple;
import net.jolsrv2.repository.neighbourhood.NaaSet;
import net.jolsrv2.repository.neighbourhood.NaaTuple;
import net.jolsrv2.repository.neighbourhood.TwoHopSet;
import net.jolsrv2.repository.neighbourhood.TwoHopTuple;

/**
 * The class implements the routing set, as specified in OLSRv2.
 * @author Andreas Schjnhaug
 *
 */
public final class RoutingSet {
	
	private static Logger logger = Logger.getLogger(RoutingSet.class.toString());
	
	
	private static RoutingSet ref;
	
	
	private ArrayList<RoutingTuple> tuples;
	
	
	private native void CreateEntry(String destination, String netmask, String gateway, String iface, int metric);
	private native void DeleteEntry(String destination);
	
	static {
        System.loadLibrary("jolsrv2_rt");
    }
	
	/**
	 * Constructor
	 *
	 */
	private RoutingSet(){
		tuples = new ArrayList<RoutingTuple>();
	}
	
	/**
	 * Implementation of the singleton design pattern
	 * @return the routing set
	 */
	public static synchronized RoutingSet getRoutingSet(){
		if(ref == null)
			ref = new RoutingSet();
		return ref;
	}
	
	/**
	 * Implementation of the singleton design pattern. This method is used to
	 * ensure that the class can not be cloned.
	 */
	public Object clone() throws CloneNotSupportedException{
		throw new CloneNotSupportedException();
	}
	
	/**
	 * Calculates the routing set as suggested in the implementation
	 *
	 */
	public void calculate(){
		
		logger.finer("Calcualting routing set"); 
		
		//Step 1
		tuples.clear();
		
		//Step 2
		ArrayList<LinkTuple> symmetricNeighbours = LinkSet.getLinkSet().getTuples(Values.SYMMETRIC);
		for(LinkTuple tuple : symmetricNeighbours){
			tuples.add(new RoutingTuple(tuple.getNeighbourCAddress().getAddress(),tuple.getNeighbourCAddress().getAddress(),1,tuple.getLocalCAddress().getAddress()));
		}
		
		//Step 3
		ArrayList<NaaTuple> naaTuples = NaaSet.getNaaSet().getTuples();
		for(NaaTuple naaTuple : naaTuples){
			if(naaTuple.getCAddresses().size() > 1){
				for(CidrAddress a1 : naaTuple.getCAddresses()){
					for(CidrAddress a2 : naaTuple.getCAddresses()){
						if(contains(a1.getAddress()) && !contains(a2.getAddress())){
							RoutingTuple a1Tuple = getTuple(a1.getAddress());
							tuples.add(new RoutingTuple(a2.getAddress(), a1Tuple.getNext(), a1Tuple.getDistance(), a1Tuple.getLocalIfaceAddr()));
						}
					}
				}
			}
		}
		
		
		//Step 4
		ArrayList<TwoHopTuple> twoHopTuples = TwoHopSet.getTwoHopSet().getTuples();
		TreeSet<LinkTuple> linkTuples = LinkSet.getLinkSet().getTuples();
		
		
		
		//TODO: also exists in mprset.populate
		//Removing interfaces in n2 which are already in n
		for(Iterator<TwoHopTuple> i = twoHopTuples.iterator(); i.hasNext();){
			TwoHopTuple t2 = i.next();
			for(LinkTuple t1 : linkTuples){
				if(t1.getNeighbourCAddress().getAddress().equals(t2.getTwoHopCAddress().getAddress()))
					i.remove();
					
					
			}
		}
		
		
		
		for(TwoHopTuple twoHopTuple : twoHopTuples){
			for(LinkTuple linkTuple : linkTuples){
				if(twoHopTuple.getNeighbourCAddress().getAddress().equals(linkTuple.getNeighbourCAddress().getAddress()) && !linkTuple.getWillingness().equals(Values.WILL_NEVER)){
					
					
					RoutingTuple r = null;
					for(RoutingTuple routingTuple : tuples){
						
						if(routingTuple.getDestination().equals(twoHopTuple.getNeighbourCAddress().getAddress())){
							r = routingTuple;
						}
					}
					if(r != null)
						tuples.add(new RoutingTuple(twoHopTuple.getTwoHopCAddress().getAddress(), r.getNext() , 2, r.getLocalIfaceAddr()));
					else
						logger.warning("ERROR 1 in RoutingSet");
				}
			}
		}
		
		
		//Step 5
		boolean cont = true;
		for(int h = 2; cont ; h++){
			cont = false;
			
			for(TopologyTuple topologyTuple : TopologySet.getTopologySet().getTuples()){
				boolean corresponds = false;
				for(RoutingTuple routingTuple : tuples){
					if(topologyTuple.getDestination().equals(routingTuple.getDestination()))
							corresponds = true;
			
				}
				if(!corresponds){
					
					RoutingTuple r = null;
					for(RoutingTuple routingTuple : tuples){
						
						if(topologyTuple.getLastHop().equals(routingTuple.getDestination()) && routingTuple.getDistance() == h){
							r = routingTuple;
						}
						
					}
					if(r != null){
						tuples.add(new RoutingTuple(topologyTuple.getDestination(),r.getNext(),h+1,r.getLocalIfaceAddr()));
						cont = true;
					}

				}
			}
		}
		
		
		try{
			
		
			for(RoutingTuple tuple : tuples){

				logger.finer(tuple.getDestination().getHostAddress());
				logger.finer(tuple.getNext().getHostAddress());
				logger.finer(tuple.getLocalIfaceAddr().getHostAddress());
				

				//TODO: Should be removed if native code is implemented for other platforms than Windows XP
				if(System.getProperty("os.name").equals("Windows XP"))
					CreateEntry(tuple.getDestination().getHostAddress(), "255.255.255.255", tuple.getNext().getHostAddress(),tuple.getLocalIfaceAddr().getHostAddress(),tuple.getDistance());
				
				
				
	
			}
		}
		catch(UnsatisfiedLinkError e){logger.warning(e.toString());}
		
		
		
		
	}
	
	/**
	 * Checks if a routing tuple with a certain destination IP address exists
	 * @param destination the destination IP address
	 * @return true if the routing tuple exists
	 */
	private boolean contains(InetAddress destination){
		for (RoutingTuple tuple : tuples){
			if(tuple.getDestination().equals(destination))
				return true;
		}
		return false;
	}
	
	/**
	 * Returns a routing tuple with a certain destination IP address if it exists
	 * @param destination the destination IP address 
	 * @return the routing tuple if exists, null if not
	 */
	private RoutingTuple getTuple(InetAddress destination){
		for(RoutingTuple tuple : tuples){
			if (tuple.getDestination().equals(destination)){
				return tuple;
			}
		}
		return null;
	}
	
	/**
	 * Deletes a routing tuple with a given destination IP address
	 * @param destination the destination IP address
	 */
	public void deleteTuple(InetAddress destination){
		try{
			for(Iterator<RoutingTuple> i = tuples.iterator(); i.hasNext(); ){
				RoutingTuple tuple = i.next();
				if (tuple.getDestination().equals(destination)){
					i.remove();
					DeleteEntry(destination.getHostAddress());
				}
				
			}
		}
		catch(Exception e){logger.warning("ERROR in GeneralSet: " + e);}
	}

}
