package net.jolsrv2.repository.topology;

import java.net.InetAddress;
import java.util.ArrayList;
import java.util.logging.Logger;

import net.jolsrv2.CidrAddress;
import net.jolsrv2.data.Tlv;

/**
 * The class implements the topology set, as specified in OLSRv2.
 * @author Andreas Schjnhaug
 *
 */


public final class TopologySet {
	private static Logger logger = Logger.getLogger(TopologySet.class.toString());
	
	private static TopologySet ref;
	
	private ArrayList<TopologyTuple> tuples;
	
	/**
	 * Constructor
	 *
	 */
	private TopologySet(){
		tuples = new ArrayList<TopologyTuple>(); 
	}
	
	/**
	 * Implementation of the singleton design pattern.
	 * @return the topology set
	 */
	public static synchronized TopologySet getTopologySet(){
		if(ref == null)
			ref = new TopologySet();
		return ref;
	}
	
	/**
	 * Implementation of the singleton design pattern. This method is used to
	 * ensure that the class can not be cloned.
	 */
	public Object clone() throws CloneNotSupportedException{
		throw new CloneNotSupportedException();
	}
	
	/**
	 * Updates the topology set as suggested in the specification
	 * @param libCAddresses the LIB CIDR address
	 * @param addresses the IP addresses
	 * @param tlvs the TLVs
	 * @param validityTime the validity time
	 * @param ansn the ANSN
	 * @return true if a change is detected
	 */
	public boolean update(ArrayList<CidrAddress> libCAddresses, ArrayList<InetAddress> addresses, ArrayList<Tlv> tlvs, long validityTime, int ansn){
				
		boolean hasChanged = false;


		for(CidrAddress libCAddress : libCAddresses){
				
			boolean found = false;
			for(InetAddress address : addresses){
				
				//TODO which has not a prefix associated (10.2)
				
				for(TopologyTuple tuple : tuples){
					if(tuple.getDestination().equals(address) && tuple.getLastHop().equals(libCAddress.getAddress())){
						found = true;
						tuple.setTime(System.nanoTime() + validityTime);
						logger.fine("- TopologySet: Updated tuple: " + tuple);
					}
				}
				if(!found){
					TopologyTuple tuple = new TopologyTuple(address, libCAddress.getAddress(), ansn);
					tuples.add(tuple);
					hasChanged = true;
					logger.fine("- TopologySet: Created tuple: " + tuple);
				}
			}
		}
		
		return hasChanged;
		
		
		
		
	}
	
	/**
	 * Returns the topology tuples
	 * @return the topology tuples
	 */
	public ArrayList<TopologyTuple> getTuples(){
		return tuples;
	}
	
}
